"""Analysis element for Quasi-Monte Carlo (QMC) sensitivity analysis.

Please refer to the article below for further references.
https://en.wikipedia.org/wiki/Variance-based_sensitivity_analysis
"""
import logging
import numpy as np
from easyvvuq import OutputType
from .base import BaseAnalysisElement
from easyvvuq.sampling import QMCSampler
from easyvvuq.sampling import MCSampler
from .ensemble_boot import confidence_interval

__author__ = 'Jalal Lakhlili'
__license__ = "LGPL"

logger = logging.getLogger(__name__)


class QMCAnalysis(BaseAnalysisElement):

    def __init__(self, sampler, qoi_cols=None):
        """Analysis element for Quasi-Monte Carlo (QMC).

        Parameters
        ----------
        sampler : easyvvuq.sampling.qmc.QMCSampler
            Sampler used to initiate the QMC analysis
        qoi_cols : list or None
            Column names for quantities of interest (for which analysis is
            performed).
        """
        if not isinstance(sampler, QMCSampler) and not isinstance(sampler, MCSampler):
            raise RuntimeError(
                'QMCAnalysis class relies on the QMCSampler or MCSampler as its sampling component')
        if qoi_cols is None:
            self.qoi_cols = list(sampler.vary.get_keys())
        else:
            self.qoi_cols = qoi_cols
        self.output_type = OutputType.SUMMARY
        self.sampler = sampler

    def element_name(self):
        """Name for this element"""
        return "QMC_Analysis"

    def element_version(self):
        """Version of this element"""
        return "0.2"

    def analyse(self, data_frame):
        """Perform QMC analysis on a given pandas DataFrame.

        Parameters
        ----------
        data_frame : pandas DataFrame
            Input data for analysis.

        Returns
        -------
        dict:
            Contains analysis results in sub-dicts with keys -
            ['statistical_moments', 'percentiles', 'sobol_indices',
             'correlation_matrices', 'output_distributions']
        """
        if data_frame.empty:
            raise RuntimeError(
                "No data in data frame passed to analyse element")

        qoi_cols = self.qoi_cols

        results = {
            'statistical_moments': {k: {} for k in qoi_cols},
            'sobols_first': {k: {} for k in qoi_cols},
            'sobols_total': {k: {} for k in qoi_cols},
            'conf_sobols_first': {k: {} for k in qoi_cols},
            'conf_sobols_total': {k: {} for k in qoi_cols}
        }

        # Extract output values for each quantity of interest from Dataframe
        samples = self.get_samples(data_frame)

        # Compute descriptive statistics for each quantity of interest
        for k in qoi_cols:
            results['statistical_moments'][k] = {'mean': np.mean(samples[k], axis=0),
                                                 'var': np.var(samples[k], axis=0),
                                                 'std': np.std(samples[k], axis=0)}
            sobols_first, conf_first, sobols_total, conf_total = \
                self.sobol_bootstrap(samples[k])
            results['sobols_first'][k] = sobols_first
            results['sobols_total'][k] = sobols_total
            results['conf_sobols_first'][k] = conf_first
            results['conf_sobols_total'][k] = conf_total

        return results

    def get_samples(self, data_frame):
        """
        Converts the Pandas dataframe into a dictionary.

        Parameters
        ----------
        data_frame : the EasyVVUQ Pandas dataframe.

        Returns
        -------
        samples : A dictionary with the QoI names as keys. Each samples[qoi_name]
        is a list of code evaluations.

        """
        samples = {k: [] for k in self.qoi_cols}
        for run_id in data_frame.run_id.unique():
            for k in self.qoi_cols:
                data = data_frame.loc[data_frame['run_id'] == run_id][k]
                samples[k].append(data.values)
        return samples

    def sobol_bootstrap(self, samples, alpha=0.05, n_samples=1000):
        """
        Computes the first order and total order Sobol indices using Saltelli's
        method. To assess the sampling inaccuracy, bootstrap confidence intervals
        are also computed.

        Reference: A. Saltelli, Making best use of model evaluations to compute
        sensitivity indices, Computer Physics Communications, 2002.

        Parameters
        ----------
        samples : list
            The samples for a given QoI.
        alpha: float
            The (1 - alpha) * 100 confidence interval parameter. The default is 0.05.
        n_samples: int
            The number of bootstrap samples. The default is 1000.

        Returns
        -------
        sobols_first_dict, conf_first_dict, sobols_total_dict, conf_total_dict:
        dictionaries containing the first- and total-order Sobol indices for all
        parameters, and (1-alpha)*100 lower and upper confidence bounds.

        """
        assert len(samples) > 0
        assert alpha > 0.0
        assert alpha < 1.0
        assert n_samples > 0

        # convert to array
        samples = np.array(samples)
        # the number of parameter and the number of MC samples in n_mc * (n_params + 2)
        # and the size of the QoI
        n_params = self.sampler.n_params
        n_mc = self.sampler.n_mc_samples
        n_qoi = samples[0].size
        sobols_first_dict = {}
        conf_first_dict = {}
        sobols_total_dict = {}
        conf_total_dict = {}

        for j, param_name in enumerate(self.sampler.vary.get_keys()):
            # code evaluations of input matrices M1, M2 and Ni, i = 1,...,n_params
            # see reference above.
            f_M2, f_M1, f_Ni = self._separate_output_values(samples, n_params, n_mc)
            # our point estimate for the 1st and total order Sobol indices
            value_first = self._first_order(f_M2, f_M1, f_Ni[:, j])
            value_total = self._total_order(f_M2, f_M1, f_Ni[:, j])
            # array for resampled estimates
            sobols_first = np.zeros([n_samples, n_qoi])
            sobols_total = np.zeros([n_samples, n_qoi])
            for i in range(n_samples):
                # resample, must be done on already seperated output due to
                # the specific order in samples
                idx = np.random.randint(0, n_mc - 1, n_mc)
                f_M2_resample = f_M2[idx]
                f_M1_resample = f_M1[idx]
                f_Ni_resample = f_Ni[idx]
                # recompute Sobol indices
                sobols_first[i] = self._first_order(f_M2_resample, f_M1_resample,
                                                    f_Ni_resample[:, j])
                sobols_total[i] = self._total_order(f_M2_resample, f_M1_resample,
                                                    f_Ni_resample[:, j])
            # compute confidence intervals
            _, low_first, high_first = confidence_interval(sobols_first, value_first,
                                                           alpha, pivotal=True)
            _, low_total, high_total = confidence_interval(sobols_total, value_total,
                                                           alpha, pivotal=True)
            # store results
            sobols_first_dict[param_name] = value_first
            conf_first_dict[param_name] = {'low': low_first, 'high': high_first}
            sobols_total_dict[param_name] = value_total
            conf_total_dict[param_name] = {'low': low_total, 'high': high_total}

        return sobols_first_dict, conf_first_dict, sobols_total_dict, conf_total_dict

    # Adapted from SALib
    @staticmethod
    def _separate_output_values(samples, n_params, n_mc_samples):
        """There are n_params + 2 different input matrices: M1, M2, N_i,
        i=1,...,n_params.  (see reference under sobol_bootstrap). The
        EasyVVUQ dataframe is stored in the order:

        [sample from M2, sample from N1, N2, ... sample from N_n_params,
         sample from M1, repeat].

        This subroutine separates the output values into the contributions
        of the different input matrices.

        Parameters
        ----------
        samples: list
            The samples for a given QoI
        n_params: int
            The number of uncertain input parameters.
        n_mc_samples: int
            The number of MC samples per input matrix, i.e. the
          number of rows in M1, M2 or Ni.

        Returns
        -------
        NumPy arrays of the separated code evaluations: f_M2, f_M1, f_Ni, where
        f_Ni contains n_params entries corresponding to the n_params Ni matrices.

        """
        evaluations = np.array(samples)

        shape = (n_mc_samples, n_params) + evaluations[0].shape
        step = n_params + 2
        f_Ni = np.zeros(shape)

        f_M2 = evaluations[0:evaluations.shape[0]:step]
        f_M1 = evaluations[(step - 1):evaluations.shape[0]:step]

        for i in range(n_params):
            f_Ni[:, i] = evaluations[(i + 1):evaluations.shape[0]:step]

        return f_M2, f_M1, f_Ni

    @staticmethod
    def _first_order(f_M2, f_M1, f_Ni):
        """Calculate first order sensitivity indices.

        Parameters
        ----------
        f_M2: NumPy array
            Array of code evaluations on input array M2
        f_M1: NumPy array
            Array of code evaluations on input array M1
        f_Ni: NumPy array
            Array of code evaluations on input array Ni, i=1,...,n_params

        Returns
        -------
        A NumPy array of the n_params first-order Sobol indices.
        """
        V = np.var(np.r_[f_M2, f_M1], axis=0)
        return np.mean(f_M1 * (f_Ni - f_M2), axis=0) / (V + (V == 0)) * (V != 0)

    @staticmethod
    def _total_order(f_M2, f_M1, f_Ni):
        """Calculate total order sensitivity indices. See also:

        A Saltelli et al, Variance based sensitivity analysis of model output.
        Design and estimator for the total sensitivity index, 2009.

        Parameters
        ----------
        f_M2: NumPy array
            Array of code evaluations on input array M2 (matrix A in ref above)
        f_M1: NumPy array
            Array of code evaluations on input array M1 (matrix B in ref above)
        f_Ni: NumPy array
            Array of code evaluations on input array Ni, i=1,...,n_params
          (matrix AB in ref above)

        Returns
        -------
        A NumPy array of the n_params total-order Sobol indices.
        """
        V = np.var(np.r_[f_M2, f_M1], axis=0)
        return 0.5 * np.mean((f_M2 - f_Ni) ** 2, axis=0) / (V + (V == 0)) * (V != 0)



import easyvvuq as uq
import chaospy as cp


def test_lhc():
    vary = {'a': cp.Uniform(-5, 3), 'b': cp.Uniform(2, 10)}
    sampler = uq.sampling.quasirandom.LHCSampler(vary, max_num=10)
    for sample in sampler:
        assert(sample['a'] >= -5.0)
        assert(sample['a'] <= 3.0)
        assert(sample['b'] >= 2.0)
        assert(sample['b'] <= 10.0)
    assert(sampler.n_samples() == 10)


def test_halton():
    vary = {'a': cp.Uniform(-5, 3), 'b': cp.Uniform(2, 10)}
    sampler = uq.sampling.quasirandom.HaltonSampler(vary, max_num=10)
    for sample in sampler:
        assert(sample['a'] >= -5.0)
        assert(sample['a'] <= 3.0)
        assert(sample['b'] >= 2.0)
        assert(sample['b'] <= 10.0)
    assert(sampler.n_samples() == 10)



"""Data structures to ensure consistency during serialization for databases.

"""
import os
import logging
import json
from easyvvuq import constants
from easyvvuq.encoders import BaseEncoder
from easyvvuq.decoders import BaseDecoder
from easyvvuq.collate import BaseCollationElement
import numpy

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"

logger = logging.getLogger(__name__)


def check_local_dir(path, dir_type='campaign'):
    """
    Check that local path exists and if not create it.

    Parameters
    ----------
    path : str
        Directory location to check.
    dir_type : str, default='campaign'
        Type of directory we are checking (used for user and debugging
        information.)

    Returns
    -------

    """

    if not os.path.isdir(path):

        if os.path.exists(path):
            logger.critical(f'{path} specified as {dir_type} directory '
                            f'for local run but is not a directory.')
            raise IOError(f'Invalid {dir_type} directory')
        else:
            os.makedirs(path)


def check_reference(ref, run_name, ref_type='campaign'):
    """
    Validation check for a `RunInfo` reference. Checks that an integer value
    has been passed to use as a reference to another 'table' - i.e. to a
    specific campaign, app or sampler.

    Parameters
    ----------
    ref : int
        Reference to be checked.
    run_name : str
        Name of run for which the check is being performed (user info/
        debugging).
    ref_type : str, default='campaign'
        Are we checking for a campaign, sampler or app (user info/
        debugging).

    Returns
    -------

    """

    if ref is None:
        message = f'No {ref_type} id specified for run {run_name}'
        logger.critical(message)
        raise RuntimeError(message)

    if not isinstance(ref, int):
        message = (f'Invalid {ref_type} id ({ref}) specified for '
                   f'run {run_name}')
        logger.critical(message)
        raise RuntimeError(message)


class RunInfo:
    """Handles information for individual application runs.

    Parameters
    ----------
    run_name : str
        Human readable name of the run.
    ensemble_name: str
        Human readable name of the ensemble this run belongs to.
    app : None or int
        ID of the associated application.
    params : None or dict
        Dictionary of parameter values for this run.
    sample: None or int
        ID of the sampler that created the run.
    campaign: None or int
        ID of the associated campaign.

    Attributes
    ----------
    campaign : int
        ID of the associated campaign.
    sample : int
        ID of the sampler that created the run.
    app : int
        ID of the associated application.
    run_name : str
        Human readable name of the run.
    ensemble_name: str
        Human readable name of the ensemble this run belongs to.
    status : enum(Status)
    """

    def __init__(
            self,
            run_name=None,
            ensemble_name=None,
            run_dir=None,
            app=None,
            params=None,
            sample=None,
            campaign=None,
            status=constants.Status.NEW):

        check_reference(campaign, run_name, ref_type='campaign')
        check_reference(sample, run_name, ref_type='sampler')
        check_reference(app, run_name, ref_type='app')

        self.campaign = campaign
        self.sample = sample
        self.app = app
        self.run_name = run_name
        self.ensemble_name = ensemble_name
        self.run_dir = run_dir

        if not params:
            message = f'No run configuration specified for run {run_name}'
            logger.critical(message)
            raise RuntimeError(message)

        self.params = params
        self.status = status

    def to_dict(self, flatten=False):
        """Convert to a dictionary (optionally flatten to single level)

        Parameters
        ----------
        flatten : bool
            Should the return dictionary be single level (i.e. should `params`
            or other dictionary variables be serialized).

        Returns
        -------
        dict
            Dictionary representing the run - if flattened then params are
            returned as a JSON format sting.
        """

        def convert_nonserializable(obj):
            if isinstance(obj, numpy.int64):
                return int(obj)
            raise TypeError('Unknown type:', type(obj))

        if flatten:

            out_dict = {
                'run_name': self.run_name,
                'ensemble_name': self.ensemble_name,
                'run_dir': self.run_dir,
                'params': json.dumps(self.params, default=convert_nonserializable),
                'status': constants.Status(self.status),
                'campaign': self.campaign,
                'sampler': self.sample,
                'app': self.app,
            }

        else:

            out_dict = {
                'run_name': self.run_name,
                'ensemble_name': self.ensemble_name,
                'run_dir': self.run_dir,
                'params': self.params,
                'status': constants.Status(self.status),
                'campaign': self.campaign,
                'sampler': self.sample,
                'app': self.app,
            }

        return out_dict


class AppInfo:
    """Handles information for particular application.

    Parameters
    ----------
    name : str or None
        Human readable application name.
    paramsspec : ParamsSpecification or None
        Description of possible parameter values.
    encoder : :obj:`easyvvuq.encoders.base.BaseEncoder`
        Encoder element for application.
    decoder : :obj:`easyvvuq.decoders.base.BaseDecoder`
        Decoder element for application.
    collater : :obj:`easyvvuq.collation.base.BaseCollationElement`
        Collater element for application.

    Attributes
    ----------
    name : str or None
        Human readable application name.
    paramsspec : ParamsSpecification or None
        Description of possible parameter values.
    input_encoder : :obj:`easyvvuq.encoders.base.BaseEncoder`
        Encoder element for application.
    output_decoder : :obj:`easyvvuq.decoders.base.BaseDecoder`
        Decoder element for application.
    collater : :obj:`easyvvuq.collation.base.BaseCollationElement`
        Collater element for application.
    """

    def __init__(
            self,
            name=None,
            paramsspec=None,
            encoder=None,
            decoder=None,
            collater=None):

        self.name = name
        self.input_encoder = encoder
        self.output_decoder = decoder
        self.collater = collater
        self.paramsspec = paramsspec

    @property
    def input_encoder(self):
        return self._input_encoder

    @input_encoder.setter
    def input_encoder(self, encoder):
        if not isinstance(encoder, BaseEncoder):
            msg = f"Provided 'encoder' must be derived from type BaseEncoder"
            logger.error(msg)
            raise Exception(msg)

        self._input_encoder = encoder

    @property
    def output_decoder(self):
        return self._output_decoder

    @output_decoder.setter
    def output_decoder(self, decoder):
        if not isinstance(decoder, BaseDecoder):
            msg = f"Provided 'decoder' must be derived from type BaseDecoder"
            logger.error(msg)
            raise Exception(msg)

        self._output_decoder = decoder

    @property
    def collater(self):
        return self._collater

    @collater.setter
    def collater(self, collater):
        if collater is None:
            msg = "A 'collater' must be provided for this App (cannot be None)."
            logging.error(msg)
            raise Exception(msg)

        if not isinstance(collater, BaseCollationElement):
            msg = "Provided 'collater' must be derived from type BaseCollationElement"
            logging.error(msg)
            raise Exception(msg)

        self._collater = collater

    def to_dict(self, flatten=False):
        """Convert to a dictionary (optionally flatten to single level)

        Parameters
        ----------
        flatten : bool
            Should the return dictionary be single level (i.e. should `paramsspec`
            be serialized).

        Returns
        -------
        dict
            Dictionary representing the application- if flattened then `paramsspec`
            is returned as a JSON format sting.
        """

        if flatten:

            out_dict = self.to_dict()

            out_dict['params'] = self.paramsspec.serialize()
        else:

            out_dict = {
                'name': self.name,
                'params': self.paramsspec,
                'input_encoder': self.input_encoder.serialize(),
                'output_decoder': self.output_decoder.serialize(),
                'collater': self.collater.serialize()
            }

        return out_dict


class CampaignInfo:
    """Handles information on Campaign.

    Parameters
    ----------
    name : str or None
        Human readable campaign name.
    easyvvuq_version : str or None
        Version of EasyVVUQ used to create the campaign.
    campaign_dir_prefix : str or None
        Prefix test for campaign directory.
    campaign_dir : str or None,
        Path to the campaign directory.
    runs_dir : str or None
        path to run directory (within the campaign directory)
    local : bool, default=False
        Is this campaign designed to be created and executed on the same
        machine?

    Attributes
    ----------
    name : str or None
        Human readable campaign name.
    easyvvuq_version : str or None
        Version of EasyVVUQ used to create the campaign.
    campaign_dir_prefix : str or None
        Prefix test for campaign directory.
    campaign_dir : str or None,
        Path to the campaign directory.
    runs_dir : str or None
        path to run directory (within the campaign directory)
    """

    def __init__(self, name=None, easyvvuq_version=None,
                 campaign_dir_prefix=None, campaign_dir=None,
                 runs_dir=None, local=False):

        if name is None:
            message = "CampaignInfo constructor must be passed a 'name'."
            logger.critical(message)
            raise RuntimeError(message)

        if campaign_dir is None:
            message = "CampaignInfo constructor must be passed 'campaign_dir'"
            logger.critical(message)
            raise RuntimeError(message)

        self.name = name
        self.campaign_dir_prefix = campaign_dir_prefix

        self.easyvvuq_version = easyvvuq_version

        # TODO: think about right location for path check for remote runs
        if local:
            check_local_dir(campaign_dir)

        self.campaign_dir = campaign_dir

        if runs_dir is None:
            runs_dir = os.path.join(campaign_dir, 'runs')

        if local:
            check_local_dir(runs_dir, 'runs')

        self.runs_dir = runs_dir

    @property
    def easyvvuq_version(self):
        return self._easyvvuq_version

    @easyvvuq_version.setter
    def easyvvuq_version(self, version_no):
        # TODO: check validity and compatibility
        self._easyvvuq_version = version_no

    def to_dict(self, flatten=False):
        """Convert this to a dictionary

        Parameters
        ----------
        flatten : bool
            Should the return dictionary be single level (always true here).

        Returns
        -------
        dict
            Dictionary representing the campaign.
        """

        out_dict = {
            'name': self.name,
            'campaign_dir': self.campaign_dir,
            'campaign_dir_prefix': self.campaign_dir_prefix,
            'runs_dir': self.runs_dir,
            'easyvvuq_version': self.easyvvuq_version
        }

        return out_dict



"""Validation by comparing QoI distributions.
"""
import numpy as np
import scipy.stats as st
from . import BaseComparisonElement


__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__author__ = 'Jalal Lakhlili'
__license__ = "LGPL"


class ValidateSimilarity(BaseComparisonElement):

    def __init__(self):
        pass

    def dist(self, p, q):
        raise NotImplementedError

    def compare(self, dataframe1, dataframe2):
        """Perform comparison between two lists or arrays
        of discrete distributions.

        Parameters
        ----------
        dataframe1 : NumPy array or list
        dataframe2 : NumPy array or list

        Returns
        -------
        A list of distances between two lists of discrete distributions,
        dataframe1 and dataframe2.
        """

        if len(dataframe1) != len(dataframe2):
            raise RuntimeError("Input dataframe sizes are not equal")

        shape = np.shape(dataframe1)
        if len(shape) == 2:
            results = []
            for i in range(len(dataframe1)):
                p1 = np.array(dataframe1[i])
                p2 = np.array(dataframe2[i])
                d = self.dist(p1, p2)
                results.append(d)
        else:
            p1 = np.array(dataframe1)
            p2 = np.array(dataframe2)
            results = self.dist(p1, p2)

        return results


class ValidateSimilarityHellinger(ValidateSimilarity):
    def element_name(self):
        return "validate_similarity_hellinger"

    def element_version(self):
        return "0.1"

    def dist(self, p, q):
        """ Compute Hellinger distance between two discrete probability
        distributions (PDF). The Hellinger distance metric gives an
        output in the range [0,1] with values closer to 0 meaning the
        PDFs are more similar.

        Parameters
        ----------
        p : NumPy array
        q : NumPy array

        Returns
        -------
        Hellinger distance between distributions p and q.
        https://en.wikipedia.org/wiki/Hellinger_distance
        """
        p /= p.sum()
        q /= q.sum()
        return np.sqrt(1. - np.sqrt(p * q).sum())


class ValidateSimilarityJensenShannon(ValidateSimilarity):
    def element_name(self):
        return "validate_similarity_jensen_shannon"

    def element_version(self):
        return "0.1"

    def dist(self, p, q):
        """ Compute Jensen-Shannon distance between two discrete
        probability distributions (PDF). It is based on Kullback–Leibler
        divergence and gives an output metric un the range [0,1] with
        values closer to 0 meaning the PDFs are more similar.

        Parameters
        ----------
        p : NumPy array
        q : NumPy array

        Returns
        -------
        Jensen-Shannon divergence between distributions p and q.
        https://en.wikipedia.org/wiki/Jensen%E2%80%93Shannon_divergence
        https://en.wikipedia.org/wiki/Kullback%E2%80%93Leibler_divergence
        """
        p /= p.sum()
        q /= q.sum()
        m = 0.5 * (p + q)
        div = 0.5 * (st.entropy(p, m) + st.entropy(q, m))
        return np.sqrt(div / np.log(2))


class ValidateSimilarityWasserstein(ValidateSimilarity):
    def element_name(self):
        return "validate_similarity_wasserstein"

    def element_version(self):
        return "0.1"

    def dist(self, p, q):
        """ Compute Wasserstein distance between two discrete cumulative
        distributions (CDF). The Wasserstein distance has an
        unrestricted range with a lower limit of 0. A smaller distance
        indicates a stronger similarity between between CFDs.

        Parameters
        ----------
        p : NumPy array
        q : NumPy array

        Returns
        -------
        Wasserstein distance between distributions p and q.
        https://en.wikipedia.org/wiki/Wasserstein_metric
        """
        return st.wasserstein_distance(p, q)



from . import RandomSampler
from .base import Vary
import logging
import numpy as np
import chaospy as cp

__author__ = "Wouter Edeling"
"""
    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


class MCSampler(RandomSampler, sampler_name='mc_sampler'):
    """
    This is a Monte Carlo sampler, used to compute the Sobol indices, mean
    and variance of the different QoIs.
    """

    def __init__(self, vary, n_mc_samples, **kwargs):
        """

        Parameters
        ----------
        + vary : a dictionary of chaospy input distributions
        + n_mc_samples : the number of MC samples. The total number of MC samples
          required to compute the Sobol indices using a Saltelli sampling plan
          will be n_mc_samples * (n_params + 1), where n_params is the number of
          uncertain parameters in vary.

        Returns
        -------
        None.

        """
        super().__init__(vary=vary, max_num=n_mc_samples, **kwargs)
        # the number of uncertain inputs
        self.n_params = len(vary)
        # the number of MC samples, for each of the n_params + 2 input matrices
        self.n_mc_samples = n_mc_samples
        self.vary = Vary(vary)
        self.count = 0
        # joint distribution
        self.joint = cp.J(*list(vary.values()))
        # create the Saltelli sampling plan
        self.saltelli(n_mc_samples)

    def __next__(self):

        if self.is_finite():
            if self.count >= self.max_num:
                raise StopIteration

        run_dict = {}
        for idx, param_name in enumerate(self.vary.get_keys()):
            run_dict[param_name] = self.xi_mc[self.count][idx]
        self.count += 1

        return run_dict

    def saltelli(self, n_mc):
        """
        Generates a Saltelli sampling plan of n_mc*(n_params + 2) input samples
        needed to compute the Sobol indices. Stored in xi_mc.

        Method: A. Saltelli, Making best use of model evaluations to compute
        sensitivity indices, Computer Physics Communications, 2002.

        Parameters
        ----------
        n_mc : the number of Monte Carlo samples per input matrix. The total
        number of samples is n_mc*(n_params + 2)

        Returns
        -------
        None.

        """
        logging.debug('Drawing input samples for Sobol index computation.')
        # the number of MC samples required to compute the Sobol indices
        self.max_num = n_mc * (self.n_params + 2)
        logging.debug('Generating {} input samples spread over {} sample matrices.'.format(
            self.max_num, self.n_params + 2))
        # Matrix M1, the sample matrix
        M_1 = self.joint.sample(n_mc).T
        # Matrix M2, the resample matrix (see reference above)
        M_2 = self.joint.sample(n_mc).T
        # array which contains all samples
        self.xi_mc = np.zeros([self.max_num, self.n_params])
        # The order in which the inputs samples must be stored is
        # [M2_1 N1_1, ..., Nd_1, M1_1, M2_2, N1_2, ...Nd_2, M1_2, M1_3 etc]
        # number of different sampling matrices
        step = self.n_params + 2
        # store M2 first, with entries separated by step places
        self.xi_mc[0:self.max_num:step] = M_2
        # store M1 entries last
        self.xi_mc[(step - 1):self.max_num:step] = M_1
        # store N_i entries between M2 and M1
        for i in range(self.n_params):
            N_i = np.array(M_2)
            # N_i = M2 with i-th colum from M1
            N_i[:, i] = M_1[:, i]
            self.xi_mc[(i + 1):self.max_num:step] = N_i
        logging.debug('Done.')

    def get_restart_dict(self):
        return {"vary": self.vary.serialize(), "n_mc_samples": self.n_mc_samples}



"""Provides element to execute a shell command in a given directory.
"""

import os
import sys
import logging
import subprocess
from . import BaseAction

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"

logger = logging.getLogger(__name__)


class ActionStatusLocal():
    def __init__(self):
        pass

    def start(self):
        return None

    def finished(self):
        return True

    def finalise(self):
        return None

    def succeeded(self):
        return True


class ExecuteLocal(BaseAction):

    def __init__(self, run_cmd, interpret=None):
        """
        Provides an action element to run a shell command in a specified
        directory.

        Parameters
        ----------

        run_cmd : str
            Command to execute.
        interpret : str or None
            Interpreter to use to execute cmd.

        """

        if os.name == 'nt':
            msg = ('Local execution is provided for testing on Posix systems'
                   'only. We detect you are using Windows.')
            logger.error(msg)
            raise NotImplementedError(msg)

        # Need to expand users, get absolute path and dereference symlinks
        self.run_cmd = os.path.realpath(os.path.expanduser(run_cmd))
        self.interpreter = interpret

    def act_on_dir(self, target_dir):
        """
        Executes `self.run_cmd` in the shell in `target_dir`.

        target_dir : str
            Directory in which to execute command.

        """

        if self.interpreter is None:
            full_cmd = f'cd "{target_dir}"\n{self.run_cmd}\n'
        else:
            full_cmd = f'cd "{target_dir}"\n{self.interpreter} {self.run_cmd}\n'
        result = os.system(full_cmd)
        if result != 0:
            sys.exit(f'Non-zero exit code from command "{full_cmd}"\n')
        return ActionStatusLocal()


class ActionStatusLocalV2():
    def __init__(self, full_cmd, target_dir):
        self.full_cmd = full_cmd
        self.target_dir = target_dir
        self.popen_object = None
        self.ret = None
        self._started = False

    def start(self):
        self.popen_object = subprocess.Popen(self.full_cmd, cwd=self.target_dir)
        self._started = True

    def started(self):
        return self._started

    def finished(self):
        """Returns true if action is finished. In this case if calling poll on
        the popen object returns a non-None value.
        """
        if self.popen_object is None:
            return False
        ret = self.popen_object.poll()
        if ret is not None:
            self.ret = ret
            return True
        else:
            return False

    def finalise(self):
        """Performs clean-up if necessary. In this case it isn't. I think.
        """
        return None

    def succeeded(self):
        """Will return True if the process finished successfully.
        It judges based on the return code and will return False
        if that code is not zero.
        """
        if not self.started():
            return False
        if self.ret != 0:
            return False
        else:
            return True


class ExecuteLocalV2(ExecuteLocal):
    """An improvement over ExecuteLocal that uses Popen and provides the non-blocking
    execution that allows you to track progress. In line with other Action classes in EasyVVUQ.
    """

    def act_on_dir(self, target_dir):
        """
        Executes `self.run_cmd` in the shell in `target_dir`.

        target_dir : str
            Directory in which to execute command.

        """

        if self.interpreter is None:
            full_cmd = self.run_cmd.split()
        else:
            full_cmd = [self.interpreter] + self.run_cmd.split()
        return ActionStatusLocalV2(full_cmd, target_dir)



import chaospy as cp
import easyvvuq as uq
import pytest
import yaml
import pickle
import numpy as np

# No longer required
# def test_l_n_exception():
#     vary = {
#         "gravity": cp.Uniform(9.8, 1.0),
#         "mass": cp.Uniform(2.0, 10.0),
#     }
#     with pytest.raises(RuntimeError):
#         sampler = uq.sampling.SCSampler(vary=vary, polynomial_order=1, sparse=True)


def test_lagrange_poly():
    assert(uq.analysis.sc_analysis.lagrange_poly(2.0, [8, 4, 9], 0) == -3.5)
    assert(uq.analysis.sc_analysis.lagrange_poly(2.0, [8, 4, 9], 1) == 2.0999999999999996)
    assert(uq.analysis.sc_analysis.lagrange_poly(2.0, [8, 4, 9], 2) == 2.4000000000000004)
    with pytest.raises(IndexError):
        uq.analysis.sc_analysis.lagrange_poly(2.0, [8, 4, 9], 3)


def test_compute_marginal():
    files = ['tests/sc/compute_marginal/0b7b64b4dcea4d5797ce270c685408a7.yml',
             'tests/sc/compute_marginal/a2676d2279224b4f91ffb39501be165d.yml',
             'tests/sc/compute_marginal/f480d6dcf4334471937bf64209c21b7d.yml']
    data = []
    for file_ in files:
        with open(file_, 'r') as fd:
            data.append(yaml.load(fd, Loader=yaml.Loader))
    analysis = pickle.load(open('tests/sc/compute_marginal/analysis.p', 'rb'))
    for datum in data:
        h, wi_d_u = analysis.compute_marginal(datum['qoi'], datum['u'],
                                              datum['u_prime'], datum['diff'])
        for index, row in enumerate(h):
            assert((row == datum['h'][index]).all())
        assert((wi_d_u == datum['wi_d_u']).all())


def test_get_sobol_indices():
    file_ = 'tests/sc/sobol_indices/6e1252d0b17f4a328c1320a390135c6a.yml'
    with open(file_, 'r') as fd:
        data = yaml.load(fd, Loader=yaml.Loader)
    analysis = pickle.load(open('tests/sc/compute_marginal/analysis.p', 'rb'))
    sobol = analysis.get_sobol_indices(data['qoi'], data['typ'])
    for key in sobol:
        assert(((sobol[key] == data['sobol'][key]) |
                (np.isnan(sobol[key]) & np.isnan(data['sobol'][key]))).all())



from .base import BaseSamplingElement, Vary

__copyright__ = """
    Copyright 2018 Robin A. Richardson, David W. Wright
    This file is part of EasyVVUQ
    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.
    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
__license__ = "LGPL"


class RandomSampler(BaseSamplingElement, sampler_name="random_sampler"):

    def __init__(self, vary=None, count=0, max_num=0):
        """
            Expects dict of var names, and their corresponding distributions
        """
        self.vary = Vary(vary)
        self.count = count
        self.max_num = max_num

    def element_version(self):
        return "0.1"

    def is_finite(self):
        if self.max_num > 0:
            return True
        return False

    def n_samples(self):
        """Returns the number of samples in this sampler.
        Returns
        -------
        if the user specifies maximum number of samples than return that, otherwise - error
        """
        if self.is_finite():
            return self.max_num
        else:
            raise RuntimeError("You can't get the number of samples in an infinite sampler")

    def __next__(self):

        if self.is_finite():
            if self.count >= self.max_num:
                raise StopIteration

        run_dict = {}
        for param_name, dist in self.vary.get_items():
            run_dict[param_name] = dist.sample(1)[0]

        self.count += 1
        return run_dict

    def is_restartable(self):
        return True

    def get_restart_dict(self):
        return {"vary": self.vary.serialize(), "max_num": self.max_num, "count": self.count}



import pytest
from unittest.mock import MagicMock, patch
import easyvvuq.actions.execute_slurm as execute_slurm
from easyvvuq.actions.execute_slurm import ExecuteSLURM, ActionStatusSLURM
import os


BATCH_FILE = """#!/bin/bash
#
#SBATCH --job-name=test
#SBATCH --output=res.txt
#
#SBATCH --ntasks=1
#SBATCH --time=10:00
#SBATCH --mem-per-cpu=100

srun python3 /EasyVVUQ/docs/epidemic/epidemic.py docs/epidemic/epidemic_in.json out.csv"""

SQUEUE_OUTPUT_1 = """CLUSTER: mpp3
             JOBID PARTITION     NAME     USER ST       TIME  NODES NODELIST(REASON)
"""

SQUEUE_OUTPUT_2 = """JOBID PARTITION     NAME     USER  ST       TIME  NODES NODELIST(REASON)
  65541 general-c hello_te      cdc  PD       0:00      2 (Priority)
"""


@patch.object(execute_slurm.subprocess, 'run')
def test_action_status_slurm(mock_subprocess_run):
    slurm_result = MagicMock()
    slurm_result.stdout.decode.return_value = "sbatch: Submitted batch job 65541"
    mock_subprocess_run.return_value = slurm_result

    action = ExecuteSLURM('tutorials/epidemic/example.slurm', '$target_dir')
    status = action.act_on_dir('.')
    assert(isinstance(status, ActionStatusSLURM))
    assert(not status.finished())
    status.start()
    assert(status.job_id == '65541')
    assert(status.started())
    slurm_result.stdout.decode.return_value = SQUEUE_OUTPUT_2
    assert(not status.finished())
    slurm_result.stdout.decode.return_value = SQUEUE_OUTPUT_1
    assert(status.finished())



from easyvvuq import Campaign
import dask.bag
from easyvvuq.constants import Status


class CampaignDask(Campaign):
    """ This extends the standard Campaign to enable execution on clusters using
    the Dask JobQueue functionality.
    """

    def apply_for_each_run_dir(self, action, client, status=Status.ENCODED):
        """
        For each run in this Campaign's run list, apply the specified action
        (an object of type Action)

        Parameters
        ----------
        action : the action to be applied to each run directory
            The function to be applied to each run directory. func() will
            be called with the run directory path as its only argument.
        client : a Dask client associated with a cluster you want to
            run your jobs on.

        Returns
        -------
        """
        run_dirs = []
        for run_id, run_data in self.campaign_db.runs(status=status, app_id=self._active_app['id']):
            run_dirs.append(run_data['run_dir'])
        bag = dask.bag.from_sequence(run_dirs)
        future = client.compute(bag.map(action.act_on_dir))
        future.result()



import sys
from .constants import OutputType
from . import data_structs
from .params_specification import ParamsSpecification
from .campaign import Campaign
from .campaign_dask import CampaignDask
from .worker import Worker
from . import actions
from . import encoders
from . import decoders
from .base_element import BaseElement
from . import sampling
from . import analysis
from . import comparison

# First make sure python version is 3.6+
assert sys.version_info >= (3, 6), (f"Python version must be >= 3.6,"
                                    f"found {sys.version_info}")

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



from ast import literal_eval


def multi_index_tuple_parser(lst):
    """
    Parses a list of strings to tuples if they represent tuples, otherwise
    leaves them as is.

    Parameters
    __________
    lst : list of strings

    Returns
    _______
    a tuple consisting of a list of tuples and/or strings and a bool indicating if
    the lst contains any tuples
    """
    contains_tuples = True
    if len(lst) == 0:
        raise RuntimeError('multi_index_tuple_parser needs a non-empty list of strings')
    for name in lst:
        if not isinstance(name, str):
            raise RuntimeError('multi_index_tuple_parser needs a list of strings')
    result = []
    for name in lst:
        if name.strip()[0] == '(' and name.strip()[-1] == ')':
            name = literal_eval(name)
        else:
            contains_tuples = False
        result.append(name)
    return result, contains_tuples



import easyvvuq as uq
import chaospy as cp
import os
import sys
import pytest
import logging
from pprint import pformat, pprint

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


# If cannonsim has not been built (to do so, run the Makefile in tests/cannonsim/src/)
# then skip this test
if not os.path.exists("tests/cannonsim/bin/cannonsim"):
    pytest.skip(
        "Skipping cannonsim test (cannonsim is not installed in tests/cannonsim/bin/)",
        allow_module_level=True)

cannonsim_path = os.path.realpath(os.path.expanduser("tests/cannonsim/bin/cannonsim"))

logging.basicConfig(level=logging.CRITICAL)


def test_recollate(tmpdir):

    num_samples = 10
    ignore_list = ['Run_1', 'Run_5', 'Run_10']

    # Define parameter space for the cannonsim app
    params = {
        "angle": {
            "type": "float",
            "min": 0.0,
            "max": 6.28,
            "default": 0.79},
        "air_resistance": {
            "type": "float",
            "min": 0.0,
            "max": 1.0,
            "default": 0.2},
        "height": {
            "type": "float",
            "min": 0.0,
            "max": 1000.0,
            "default": 1.0},
        "time_step": {
            "type": "float",
            "min": 0.0001,
            "max": 1.0,
            "default": 0.01},
        "gravity": {
            "type": "float",
            "min": 0.0,
            "max": 1000.0,
            "default": 9.8},
        "mass": {
            "type": "float",
            "min": 0.0001,
            "max": 1000.0,
            "default": 1.0},
        "velocity": {
            "type": "float",
            "min": 0.0,
            "max": 1000.0,
            "default": 10.0}}

    # Create an encoder and decoder for the cannonsim app
    encoder = uq.encoders.GenericEncoder(
        template_fname='tests/cannonsim/test_input/cannonsim.template',
        delimiter='#',
        target_filename='in.cannon')
    output_cols = ['Dist', 'lastvx', 'lastvy']
    decoder = uq.decoders.SimpleCSV(
        target_filename='output.csv', output_columns=output_cols, header=0)
    collater = uq.collate.AggregateSamples(average=False)

    # Set up samplers
    vary = {
        "gravity": cp.Uniform(9.8, 1.0),
        "mass": cp.Uniform(2.0, 10.0),
    }
    sampler = uq.sampling.RandomSampler(vary=vary, max_num=num_samples)

    my_campaign = uq.Campaign(name='test', work_dir=tmpdir, db_location='sqlite:///')
    my_campaign.add_app(name="cannon",
                        params=params,
                        encoder=encoder,
                        decoder=decoder,
                        collater=collater)
    my_campaign.set_app("cannon")

    my_campaign.set_sampler(sampler)
    my_campaign.draw_samples()
    my_campaign.populate_runs_dir()

    actions = uq.actions.ExecuteLocal("tests/cannonsim/bin/cannonsim in.cannon output.csv")
    my_campaign.apply_for_each_run_dir(actions)
    my_campaign.collate()

    # Set some runs to be IGNORED, then recollate all
    my_campaign.ignore_runs(ignore_list)
    my_campaign.recollate()

    # Check that the right number of rows are in the collation dataframe
    assert(len(my_campaign.get_collation_result().index) == num_samples - len(ignore_list))

    # Rerun some runs
    my_campaign.rerun(['Run_2', 'Run_3', 'Run_4'])
    my_campaign.apply_for_each_run_dir(actions)

    pprint(my_campaign._log)


if __name__ == "__main__":
    test_recollate('/tmp/')



from easyvvuq.sampling.sampler_of_samplers import MultiSampler
from easyvvuq.sampling.sweep import BasicSweep
import pytest
import unittest.mock as mock


@pytest.fixture
def multi_sampler():
    sampler1 = mock.MagicMock()
    sampler2 = mock.MagicMock()
    sampler1.is_finite.return_value = True
    sampler2.is_finite.return_value = True
    multi_sampler = MultiSampler(sampler1, sampler2)
    return multi_sampler


def test_init_exceptions():
    # All samplers must be finite
    sampler1 = mock.MagicMock()
    sampler2 = mock.MagicMock()
    sampler1.is_finite.return_value = True
    sampler2.is_finite.return_value = False
    with pytest.raises(RuntimeError):
        multi_sampler = MultiSampler(sampler1, sampler2)
    with pytest.raises(RuntimeError):
        MultiSampler()


def test_element_version(multi_sampler):
    assert(isinstance(multi_sampler.element_version(), str))


def test_is_finite(multi_sampler):
    assert(multi_sampler.is_finite())


def test_n_samples():
    sampler1 = BasicSweep({'a': [1, 2, 3], 'b': [4, 5, 6]})
    sampler2 = BasicSweep({'a': [1, 2, 3], 'b': [4, 5, 6]})
    multi = MultiSampler(sampler1, sampler2)
    assert(multi.n_samples() == 81)


def test_iterator():
    def popper(values):
        if len(values):
            values.pop()
        else:
            raise StopIteration
    sampler1 = mock.MagicMock()
    values1 = [3, 2, 1]
    sampler1.__next__.return_value = values1[-1]
    sampler1.__next__.side_effect = popper(values1)
    sampler2 = mock.MagicMock()
    values2 = [3, 2, 1]
    sampler2.__next__.return_value = values2[-1]
    sampler2.__next__.side_effect = popper(values2)
    sampler1.is_finite.return_value = True
    sampler2.is_finite.return_value = True


def test_is_restartable(multi_sampler):
    assert(multi_sampler.is_restartable())


def test_get_restart_dict(multi_sampler):
    assert(isinstance(multi_sampler.get_restart_dict(), dict))



# for the Sobol g function, the exact (1st-order)
# Sobol indices are known analytically
import matplotlib.pyplot as plt
import os
import easyvvuq as uq
import numpy as np
import chaospy as cp


def print_exact_sobols():
    V_i = np.zeros(d)

    for i in range(d):
        V_i[i] = 1.0 / (3.0 * (1 + a[i])**2)

    V = np.prod(1 + V_i) - 1

    print('----------------------')
    print('Exact 1st-order Sobol indices: ', V_i / V)


plt.close('all')

# number of unknown variables
d = 5

# parameters required by test function
a = [0.0, 0.5, 3.0, 9.0, 99.0]

# author: Wouter Edeling
__license__ = "LGPL"

# home directory of user
home = os.path.expanduser('~')
HOME = os.path.abspath(os.path.dirname(__file__))

# Set up a fresh campaign called "sc"
my_campaign = uq.Campaign(name='sc', work_dir='/tmp')

# Define parameter space
params = {
    "x1": {
        "type": "float",
        "min": 0.0,
        "max": 1.0,
        "default": 0.5},
    "x2": {
        "type": "float",
        "min": 0.0,
        "max": 1.0,
        "default": 0.5},
    "x3": {
        "type": "float",
        "min": 0.0,
        "max": 1.0,
        "default": 0.5},
    "x4": {
        "type": "float",
        "min": 0.0,
        "max": 1.0,
        "default": 0.5},
    "x5": {
        "type": "float",
        "min": 0.0,
        "max": 1.0,
        "default": 0.5},
    "out_file": {
        "type": "string",
        "default": "output.csv"}}

output_filename = params["out_file"]["default"]
output_columns = ["f"]

# Create an encoder, decoder and collation element
encoder = uq.encoders.GenericEncoder(
    template_fname=HOME + '/sobol.template',
    delimiter='$',
    target_filename='sobol_in.json')
decoder = uq.decoders.SimpleCSV(target_filename=output_filename,
                                output_columns=output_columns,
                                header=0)
collater = uq.collate.AggregateSamples(average=False)

# Add the SC app (automatically set as current app)
my_campaign.add_app(name="sc",
                    params=params,
                    encoder=encoder,
                    decoder=decoder,
                    collater=collater)

# Create the sampler
vary = {
    "x1": cp.Uniform(0.0, 1.0),
    "x2": cp.Uniform(0.0, 1.0),
    "x3": cp.Uniform(0.0, 1.0),
    "x4": cp.Uniform(0.0, 1.0),
    "x5": cp.Uniform(0.0, 1.0)}

"""
SPARSE GRID PARAMETERS
----------------------
- sparse = True: use a Smolyak sparse grid
- growth = True: use an exponential rule for the growth of the number
  of 1D collocation points per level. Used to make e.g. clenshaw-curtis
  quadrature nested.
"""
my_sampler = uq.sampling.SCSampler(vary=vary, polynomial_order=2,
                                   quadrature_rule="G", sparse=False,
                                   growth=False)

# Associate the sampler with the campaign
my_campaign.set_sampler(my_sampler)

# Will draw all (of the finite set of samples)
my_campaign.draw_samples()
my_campaign.populate_runs_dir()

# Use this instead to run the samples using EasyVVUQ on the localhost
my_campaign.apply_for_each_run_dir(uq.actions.ExecuteLocal(
    "tests/sc/sobol_model.py sobol_in.json"))

my_campaign.collate()

# Post-processing analysis
analysis = uq.analysis.SCAnalysis(sampler=my_sampler, qoi_cols=output_columns)

my_campaign.apply_analysis(analysis)

results = my_campaign.get_last_analysis()

print(results['sobols_first'])

print_exact_sobols()

plt.show()



import easyvvuq as uq
import chaospy as cp
import os
import sys
import pytest
from pprint import pprint

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


# If cannonsim has not been built (to do so, run the Makefile in tests/cannonsim/src/)
# then skip this test
if not os.path.exists("tests/cannonsim/bin/cannonsim"):
    pytest.skip(
        "Skipping cannonsim test (cannonsim is not installed in tests/cannonsim/bin/)",
        allow_module_level=True)

CANNONSIM_PATH = os.path.realpath(os.path.expanduser("tests/cannonsim/bin/cannonsim"))


def test_clear_collate(tmpdir):

    # Set up a fresh campaign called "cannon"
    my_campaign = uq.Campaign(name='cannon', work_dir=tmpdir)

    # Define parameter space for the cannonsim app
    params = {
        "angle": {
            "type": "float",
            "min": 0.0,
            "max": 6.28,
            "default": 0.79},
        "air_resistance": {
            "type": "float",
            "min": 0.0,
            "max": 1.0,
            "default": 0.2},
        "height": {
            "type": "float",
            "min": 0.0,
            "max": 1000.0,
            "default": 1.0},
        "time_step": {
            "type": "float",
            "min": 0.0001,
            "max": 1.0,
            "default": 0.01},
        "gravity": {
            "type": "float",
            "min": 0.0,
            "max": 1000.0,
            "default": 9.8},
        "mass": {
            "type": "float",
            "min": 0.0001,
            "max": 1000.0,
            "default": 1.0},
        "velocity": {
            "type": "float",
            "min": 0.0,
            "max": 1000.0,
            "default": 10.0}}

    # Create an encoder, decoder and collater for the cannonsim app
    encoder = uq.encoders.GenericEncoder(
        template_fname='tests/cannonsim/test_input/cannonsim.template',
        delimiter='#',
        target_filename='in.cannon')
    decoder = uq.decoders.SimpleCSV(
        target_filename='output.csv', output_columns=[
            'Dist', 'lastvx', 'lastvy'], header=0)
    collater = uq.collate.AggregateSamples(average=False)

    # Add the cannonsim app
    my_campaign.add_app(name="cannonsim",
                        params=params,
                        encoder=encoder,
                        decoder=decoder,
                        collater=collater)

    # Set the active app to be cannonsim (this is redundant when only one app
    # has been added)
    my_campaign.set_app("cannonsim")

    # Set up samplers
    vary = {
        "gravity": cp.Uniform(9.8, 1.0),
        "mass": cp.Uniform(2.0, 10.0),
    }
    sampler = uq.sampling.RandomSampler(vary=vary, max_num=5)

    # Set the campaign to use this sampler
    my_campaign.set_sampler(sampler)

    # Test reloading
    my_campaign.save_state(tmpdir + "test_multisampler.json")
    reloaded_campaign = uq.Campaign(state_file=tmpdir + "test_multisampler.json", work_dir=tmpdir)

    # Draw all samples
    my_campaign.draw_samples()

    # Encode
    my_campaign.populate_runs_dir()

    # Execute
    my_campaign.apply_for_each_run_dir(
        uq.actions.ExecuteLocal("tests/cannonsim/bin/cannonsim in.cannon output.csv"))

    # Collate results
    my_campaign.collate()
    print("data:", my_campaign.get_collation_result())

    # Clear collation and print results again
    my_campaign.clear_collation()
    print("data:", my_campaign.get_collation_result())

    assert(my_campaign.get_collation_result() is None)

    pprint(my_campaign._log)


if __name__ == "__main__":
    test_clear_collate("/tmp/")



"""An encoder meant to simply copy a file to the input directory unchanged.
It is meant to be used in combination with MultiBuilder encoder and possibly
the DirectoryBuilder. It duplicates some functionality of the ApplyFixtures
encoder but can be useful for very simple cases.

Examples
--------
>>> multiencoder = uq.encoders.MultiEncoder(
      DirectoryBuilder(tree={"parent" : {"child1" : None, "child2" : None}}),
      CopyEncoder('/home/user/input1.conf', 'parent/child1/input1.conf')
      CopyEncoder('/home/user/input2.conf', 'parent/child1/input2.conf')
      CopyEncoder('/home/user/input3.conf', 'parent/child2/input3.conf')
      GenericEncoder(delimiter='$', template_fname='/home/user/template.in',
                     target_filename='parent/input.int'))
"""

import os
import shutil
from .base import BaseEncoder


class CopyEncoder(BaseEncoder, encoder_name="copy_encoder"):
    """An Encoder to copy an input file to a simulation.

    Parameters
    ----------
    source_filename : str
      a full path to some file that a simulation needs

    target_filename : str
      a target filename inside the simulation directory
    """

    def __init__(self, source_filename, target_filename):
        self.source_filename = source_filename
        self.target_filename = target_filename
        if not os.path.isfile(self.source_filename):
            raise RuntimeError("Specified source file does not exist:", source_filename)

    def encode(self, params={}, target_dir=''):
        """Copy a file to target_dir.

        Parameters
        ----------
        params : dict
           keep empty, has no effect

        target_dir : str
           target directory, full path
        """
        if not os.path.isdir(target_dir):
            raise RuntimeError("Specified target directory does not exist:", target_dir)
        shutil.copyfile(self.source_filename, os.path.join(target_dir, self.target_filename))

    def get_restart_dict(self):
        return {"source_filename": self.source_filename,
                "target_filename": self.target_filename}

    def element_version(self):
        return "0.1"



import os
import easyvvuq as uq
import numpy as np
import chaospy as cp
import pytest
import logging
import pandas as pd
from tests.sc.sobol_model import sobol_g_func


def exact_sobols_g_func(d=2, a=[0.0, 0.5, 3.0, 9.0, 99.0]):
    # for the Sobol g function, the exact (1st-order)
    # Sobol indices are known analytically
    V_i = np.zeros(d)
    for i in range(d):
        V_i[i] = 1.0 / (3.0 * (1 + a[i])**2)
    V = np.prod(1 + V_i) - 1
    logging.debug('Exact 1st-order Sobol indices: ', V_i / V)
    return V_i / V


# author: Wouter Edeling
__license__ = "LGPL"

HOME = os.path.abspath(os.path.dirname(__file__))


@pytest.fixture
def data():
    # fix random seed to make this test deterministic
    np.random.seed(10000000)
    # Create the sampler
    vary = {
        "x1": cp.Uniform(0.0, 1.0),
        "x2": cp.Uniform(0.0, 1.0)
    }
    # Select the MC sampler
    mc_sampler = uq.sampling.MCSampler(vary, n_mc_samples=100)
    data = []
    for run_id, sample in enumerate(mc_sampler):
        data.append({
            'run_id': run_id,
            'x1': sample['x1'],
            'x2': sample['x2'],
            'f': sobol_g_func([sample['x1'], sample['x2']], d=2)
        })

    df = pd.DataFrame(data)
    return mc_sampler, df


@pytest.fixture
def results(data):
    # Post-processing analysis
    mc_sampler, df = data
    analysis = uq.analysis.QMCAnalysis(sampler=mc_sampler, qoi_cols=['f'])
    results = analysis.analyse(df)
    return results


def test_mc_analysis(results):
    # analytic Sobol indices
    ref_sobols = exact_sobols_g_func()
    sobol_x1 = results['sobols_first']['f']['x1']
    sobol_x2 = results['sobols_first']['f']['x2']
    assert sobol_x1 == pytest.approx(ref_sobols[0], abs=0.1)
    assert sobol_x2 == pytest.approx(ref_sobols[1], abs=0.1)


def test_sobol_bootstrap(data):
    mc_sampler, df = data
    analysis = uq.analysis.QMCAnalysis(sampler=mc_sampler, qoi_cols=['f'])
    s1, s1_conf, st, st_conf = analysis.sobol_bootstrap(list(df['f']))
    assert(s1['x1'] == pytest.approx(0.5569058947880715, 0.01))
    assert(s1['x2'] == pytest.approx(0.20727553481694053, 0.01))
    assert(st['x1'] == pytest.approx(0.8132793654841785, 0.01))
    assert(st['x2'] == pytest.approx(0.3804962894947435, 0.01))
    assert(s1_conf['x1']['low'][0] == pytest.approx(0.14387035, 0.01))
    assert(s1_conf['x1']['high'][0] == pytest.approx(0.89428774, 0.01))
    assert(s1_conf['x2']['low'][0] == pytest.approx(-0.11063341, 0.01))
    assert(s1_conf['x2']['high'][0] == pytest.approx(0.46752829, 0.01))
    assert(st_conf['x1']['low'][0] == pytest.approx(0.61368887, 0.01))
    assert(st_conf['x1']['high'][0] == pytest.approx(1.01858671, 0.01))
    assert(st_conf['x2']['low'][0] == pytest.approx(0.24361207, 0.01))
    assert(st_conf['x2']['high'][0] == pytest.approx(0.49214117, 0.01))


def test_separate_output_values(data):
    mc_sampler, df = data
    analysis = uq.analysis.QMCAnalysis(sampler=mc_sampler, qoi_cols=['f'])
    f_M2, f_M1, f_Ni = analysis._separate_output_values(df['f'], 2, 100)
    assert(f_M2.shape == (100,))
    assert(f_M1.shape == (100,))
    assert(f_Ni.shape == (100, 2))


def test_get_samples(data):
    pass



import pytest
from easyvvuq.utils.helpers import multi_index_tuple_parser


def test_multi_index_tuple_parser_exceptions():
    with pytest.raises(RuntimeError):
        multi_index_tuple_parser([])
    with pytest.raises(RuntimeError):
        multi_index_tuple_parser([3])
    with pytest.raises(RuntimeError):
        multi_index_tuple_parser(['a', 'b', 3])
    with pytest.raises(RuntimeError):
        multi_index_tuple_parser([3, 'a', 'b'])


def test_multi_index_parser_corner_cases():
    assert(multi_index_tuple_parser(["a"]) == (["a"], False))
    assert(multi_index_tuple_parser(['a', '("b", 1)']) == (["a", ("b", 1)], False))


def test_multi_index_parser():
    assert(multi_index_tuple_parser(['("a", 1)', '("b", 1)']) == ([("a", 1), ("b", 1)], True))



import pytest
import time
from unittest.mock import MagicMock
from easyvvuq.actions import ActionStatuses


def test_action_status_kubernetes():
    status1, status2, status3 = (MagicMock(), MagicMock(), MagicMock())
    status1.finished.return_value = False
    status2.finished.return_value = True
    status3.finished.return_value = True
    status1.succeeded.return_value = False
    status2.succeeded.return_value = False
    status3.succeeded.return_value = True
    statuses = ActionStatuses([status1, status2, status3], 3)
    time.sleep(1)
    stats = statuses.progress()
    assert(stats['active'] == 1)
    assert(stats['finished'] == 1)
    assert(stats['failed'] == 1)
    assert(not status1.finalise.called)
    assert(not status2.finalise.called)
    assert(status3.finalise.called)
    status1.finished.return_value = True
    status2.finished.return_value = True
    status3.finished.return_value = True
    status1.succeeded.return_value = True
    status2.succeeded.return_value = True
    status3.succeeded.return_value = True



import pytest
from unittest.mock import MagicMock
import easyvvuq.actions.execute_kubernetes as execute_kubernetes
from easyvvuq.actions.execute_kubernetes import ActionStatusKubernetes
import os


# Monkey patch some stuff

execute_kubernetes.config = MagicMock()
execute_kubernetes.core_v1_api = MagicMock()
execute_kubernetes.Configuration = MagicMock()
execute_kubernetes.V1ConfigMap = MagicMock()
execute_kubernetes.V1ObjectMeta = MagicMock()


def test_execute_kubernetes():
    action = execute_kubernetes.ExecuteKubernetes(
        'tests/kubernetes/epidemic.yaml', ['epidemic.json'], 'out.csv')
    action.act_on_dir('tests/kubernetes')


def test_action_status_kubernetes():
    api = MagicMock()
    pod_name = 'test'
    config_names = [('a', 'b'), ('c', 'd')]
    namespace = 'test_namespace'
    outfile = 'test.csv'
    status = ActionStatusKubernetes(
        api, {'metadata': {'name': 'test'}},
        config_names, namespace, outfile)
    resp = MagicMock()
    resp.status.phase = 'Pending'
    api.read_namespaced_pod.return_value = resp
    assert(not status.finished())
    with pytest.raises(RuntimeError):
        status.finalise()
    assert(not status.succeeded())
    resp.status.phase = 'Succeeded'
    assert(status.finished())
    api.read_namespaced_pod_log.return_value = 'testing'
    status.finalise()
    assert(os.path.isfile('test.csv'))
    with open('test.csv', 'r') as fd:
        assert(fd.read() == 'testing')
    os.remove('test.csv')
    assert(api.delete_namespaced_config_map.called_with('b', 'test_namespace'))
    assert(api.delete_namespaced_config_map.called_with('d', 'test_namespace'))
    assert(api.delete_namespaced_pod.called_with('test', 'test_namespace'))



import json

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


class BaseElement(object):
    """Baseclass for all EasyVVUQ elements.

    Attributes
    ----------

    """

    def element_name(self):
        raise NotImplementedError

    def element_version(self):
        raise NotImplementedError

    def element_category(self):
        raise NotImplementedError

    def is_restartable(self):
        return False

    def get_restart_dict(self):
        return None

    def serialize(self):
        return json.dumps({
            "element_name": self.element_name(),
            "element_version": self.element_version(),
            "element_category": self.element_category(),
            "restartable": self.is_restartable(),
            "state": self.get_restart_dict()
        })



"""Analysis element for polynomial chaos expansion (PCE).
"""
import logging
import chaospy as cp
from easyvvuq import OutputType
from .base import BaseAnalysisElement

__author__ = 'Jalal Lakhlili'
__license__ = "LGPL"

logger = logging.getLogger(__name__)


class PCEAnalysis(BaseAnalysisElement):

    def __init__(self, sampler=None, qoi_cols=None):
        """Analysis element for polynomial chaos expansion (PCE).

        Parameters
        ----------
        sampler : :obj:`easyvvuq.sampling.pce.PCESampler`
            Sampler used to initiate the PCE analysis
        qoi_cols : list or None
            Column names for quantities of interest (for which analysis is
            performed).
        """

        if sampler is None:
            msg = 'PCE analysis requires a paired sampler to be passed'
            raise RuntimeError(msg)

        if qoi_cols is None:
            raise RuntimeError("Analysis element requires a list of "
                               "quantities of interest (qoi)")

        self.qoi_cols = qoi_cols
        self.output_type = OutputType.SUMMARY
        self.sampler = sampler

    def element_name(self):
        """Name for this element for logging purposes"""
        return "PCE_Analysis"

    def element_version(self):
        """Version of this element for logging purposes"""
        return "0.5"

    def analyse(self, data_frame=None):
        """Perform PCE analysis on input `data_frame`.

        Parameters
        ----------
        data_frame : :obj:`pandas.DataFrame`
            Input data for analysis.

        Returns
        -------
        dict:
            Contains analysis results in sub-dicts with keys -
            ['statistical_moments', 'percentiles', 'sobol_indices',
             'correlation_matrices', 'output_distributions']
        """

        if data_frame is None:
            raise RuntimeError("Analysis element needs a data frame to "
                               "analyse")
        elif data_frame.empty:
            raise RuntimeError(
                "No data in data frame passed to analyse element")

        qoi_cols = self.qoi_cols

        results = {'statistical_moments': {},
                   'percentiles': {},
                   'sobols_first': {k: {} for k in qoi_cols},
                   'sobols_second': {k: {} for k in qoi_cols},
                   'sobols_total': {k: {} for k in qoi_cols},
                   'correlation_matrices': {},
                   'output_distributions': {},
                   }

        # Get the Polynomial
        P = self.sampler.P

        # Get the PCE variante to use (Regression or Projection)
        regression = self.sampler.regression

        # Compute nodes (and weights)
        if regression:
            nodes = cp.generate_samples(order=self.sampler.n_samples,
                                        domain=self.sampler.distribution,
                                        rule=self.sampler.rule)
        else:
            nodes, weights = cp.generate_quadrature(order=self.sampler.polynomial_order,
                                                    dist=self.sampler.distribution,
                                                    rule=self.sampler.rule,
                                                    sparse=self.sampler.quad_sparse,
                                                    growth=self.sampler.quad_growth)

        # Extract output values for each quantity of interest from Dataframe
        samples = {k: [] for k in qoi_cols}
        for run_id in data_frame.run_id.unique():
            for k in qoi_cols:
                data = data_frame.loc[data_frame['run_id'] == run_id][k]
                samples[k].append(data.values)

        # Compute descriptive statistics for each quantity of interest
        for k in qoi_cols:
            # Approximation solver
            if regression:
                if samples[k][0].dtype == object:
                    for i in range(self.sampler.count):
                        samples[k][i] = samples[k][i].astype("float64")
                fit = cp.fit_regression(P, nodes, samples[k], "T")
            else:
                fit = cp.fit_quadrature(P, nodes, weights, samples[k])

            # Statistical moments
            mean = cp.E(fit, self.sampler.distribution)
            var = cp.Var(fit, self.sampler.distribution)
            std = cp.Std(fit, self.sampler.distribution)
            results['statistical_moments'][k] = {'mean': mean,
                                                 'var': var,
                                                 'std': std}

            # Percentiles (Pxx)
            P10 = cp.Perc(fit, 10, self.sampler.distribution)
            P90 = cp.Perc(fit, 90, self.sampler.distribution)
            results['percentiles'][k] = {'p10': P10, 'p90': P90}

            # Sensitivity Analysis: First, Second and Total Sobol indices
            sobols_first_narr = cp.Sens_m(fit, self.sampler.distribution)
            sobols_second_narr = cp.Sens_m2(fit, self.sampler.distribution)
            sobols_total_narr = cp.Sens_t(fit, self.sampler.distribution)
            sobols_first_dict = {}
            sobols_second_dict = {}
            sobols_total_dict = {}
            ipar = 0
            i = 0
            for param_name in self.sampler.vary.get_keys():
                j = self.sampler.params_size[ipar]
                sobols_first_dict[param_name] = sobols_first_narr[i:i + j]
                sobols_second_dict[param_name] = sobols_second_narr[i:i + j]
                sobols_total_dict[param_name] = sobols_total_narr[i:i + j]
                i += j
                ipar += 1
            results['sobols_first'][k] = sobols_first_dict
            results['sobols_second'][k] = sobols_second_dict
            results['sobols_total'][k] = sobols_total_dict

            # Correlation matrix
            results['correlation_matrices'][k] = cp.Corr(
                fit, self.sampler.distribution)

            # Output distributions
            results['output_distributions'][k] = cp.QoI_Dist(
                fit, self.sampler.distribution)

        return results



from .base import BaseAction
from .execute_local import ExecuteLocal, ExecuteLocalV2
from .execute_kubernetes import ExecuteKubernetes
from .execute_slurm import ExecuteSLURM
from .action_statuses import ActionStatuses

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"



"""Replica Sampler

Summary
-------

Primarily intended for sampling the same paramater values but with
different random seed. Other uses may be possible. It takes a finite
sampler and produces an infinite sampler from it. This infinite
sampler loops through the parameters produced by the finite sampler
and at each cycle adds a unique id number for that cycle to the
parameter dictionary.
"""

from easyvvuq.sampling import BaseSamplingElement
from itertools import cycle


class ReplicaSampler(BaseSamplingElement, sampler_name='replica_sampler'):
    """Replica Sampler

    Parameters
    ----------
    sampler : an instance of a class derived from  BaseSamplingElement
        a finite sampler to loop over

    ensemble_col : string
        a parameter name for the ensemble id
    """

    def __init__(self, sampler, ensemble_col='ensemble'):
        if not sampler.is_finite():
            raise RuntimeError("Replica sampler only works with finite samplers")
        self.sampler = sampler
        self.ensemble_col = ensemble_col
        self.history = []
        for sample in sampler:
            self.history.append(sample)
        self.size = len(self.history)
        self.cycle = cycle(self.history)
        self.counter = 0

    def is_finite(self):
        return False

    def element_version(self):
        return '0.1'

    def n_samples(self):
        raise RuntimeError("You can't get the number of samples in an infinite sampler")

    def __next__(self):
        params = dict(next(self.cycle))
        params[self.ensemble_col] = self.counter
        self.counter = (self.counter + 1) % self.size
        return params

    def is_restartable(self):
        return False



from easyvvuq.encoders.directory_builder import DirectoryBuilder
import pytest
import os


@pytest.fixture
def builder():
    return DirectoryBuilder({'a': {'b': {'c': None, 'd': None}}, 'e': {'f': None}})


def test_init(builder):
    assert(builder.tree == {'a': {'b': {'c': None, 'd': None}}, 'e': {'f': None}})


def test_encoder(builder, tmp_path):
    with pytest.raises(RuntimeError):
        builder.encode()


def test_create_dir_tree(builder, tmp_path):
    builder.create_dir_tree(builder.tree, tmp_path)
    assert(os.path.isdir(os.path.join(tmp_path, 'a', 'b', 'c')))
    assert(os.path.isdir(os.path.join(tmp_path, 'a', 'b', 'd')))
    assert(os.path.isdir(os.path.join(tmp_path, 'e', 'f')))


def test_get_restart_dict(builder):
    restart_dict = builder.get_restart_dict()
    assert(restart_dict['tree'] == {'a': {'b': {'c': None, 'd': None}}, 'e': {'f': None}})


def test_element_version(builder):
    assert(isinstance(builder.element_version(), str))



"""Classes implementing the sampling element for EasyVVUQ

Summary
-------
Samplers in the context of EasyVVUQ are classes that generate
sequences of parameter dictionaries. These dictionaries are then used
to create input files for the simulations.
"""

from .base import BaseSamplingElement, Vary
from .random import RandomSampler
from .stochastic_collocation import SCSampler
from .pce import PCESampler
from .qmc import QMCSampler
from .sweep import BasicSweep
from .sampler_of_samplers import MultiSampler
from .quasirandom import LHCSampler
from .empty import EmptySampler
from .replica_sampler import ReplicaSampler
from .mc_sampler import MCSampler

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"



import easyvvuq as uq
import chaospy as cp
import os
import sys
import pytest
import logging
from pprint import pformat, pprint
from .gauss.encoder_gauss import GaussEncoder
from .gauss.decoder_gauss import GaussDecoder

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


# If cannonsim has not been built (to do so, run the Makefile in tests/cannonsim/src/)
# then skip this test
if not os.path.exists("tests/cannonsim/bin/cannonsim"):
    pytest.skip(
        "Skipping cannonsim test (cannonsim is not installed in tests/cannonsim/bin/)",
        allow_module_level=True)

cannonsim_path = os.path.realpath(os.path.expanduser("tests/cannonsim/bin/cannonsim"))


def execute_cannonsim(path, params):
    os.system(f"cd {path} && {cannonsim_path} in.cannon output.csv")


logging.basicConfig(level=logging.CRITICAL)


@pytest.fixture
def campaign_test():
    class CampaignTest:
        def __init__(self, campaign_name, work_dir, db_type='sql'):
            self.campaign = uq.Campaign(name=campaign_name, work_dir=work_dir, db_type=db_type)

        def run_tests(self):
            pass


@pytest.fixture
def campaign():
    def _campaign(work_dir, campaign_name, app_name, params, encoder, decoder, sampler,
                  collater, actions, stats, vary, num_samples=0, replicas=1, db_type='sql',
                  call_fn=None):
        my_campaign = uq.Campaign(name=campaign_name, work_dir=work_dir, db_type=db_type)
        logging.debug("Serialized encoder: %s", str(encoder.serialize()))
        logging.debug("Serialized decoder: %s", str(decoder.serialize()))
        logging.debug("Serialized collation: %s", str(collater.serialize()))

        # Add the cannonsim app
        my_campaign.add_app(name=app_name,
                            params=params,
                            encoder=encoder,
                            decoder=decoder,
                            collater=collater)
        my_campaign.set_app(app_name)
        logging.debug("Serialized sampler: %s", str(sampler.serialize()))
        # Set the campaign to use this sampler
        my_campaign.set_sampler(sampler)
        # Draw 5 samples
        my_campaign.draw_samples(num_samples=num_samples, replicas=replicas)
        # Print the list of runs now in the campaign db
        logging.debug("List of runs added:")
        logging.debug(pformat(my_campaign.list_runs()))
        logging.debug("---")
        # Encode all runs into a local directory
        logging.debug(pformat(
            f"Encoding all runs to campaign runs dir {my_campaign.get_campaign_runs_dir()}"))
        my_campaign.populate_runs_dir()
        assert(len(my_campaign.get_campaign_runs_dir()) > 0)
        assert(os.path.exists(my_campaign.get_campaign_runs_dir()))
        assert(os.path.isdir(my_campaign.get_campaign_runs_dir()))
        if call_fn is not None:
            my_campaign.call_for_each_run(call_fn)
        # Local execution
        if actions is not None:
            my_campaign.apply_for_each_run_dir(actions)
        # Collate all data into one pandas data frame
        my_campaign.collate()
        logging.debug("data: %s", str(my_campaign.get_collation_result()))
        # Save the state of the campaign
        state_file = work_dir + "{}_state.json".format(app_name)
        my_campaign.save_state(state_file)
        my_campaign = None
        # Load state in new campaign object
        reloaded_campaign = uq.Campaign(state_file=state_file, work_dir=work_dir)
        reloaded_campaign.set_app(app_name)
        # Draw 3 more samples, execute, and collate onto existing dataframe
        logging.debug("Running 3 more samples...")
        reloaded_campaign.draw_samples(num_samples=num_samples, replicas=replicas)
        logging.debug("List of runs added:")
        logging.debug(pformat(reloaded_campaign.list_runs()))
        logging.debug("---")
        reloaded_campaign.populate_runs_dir()
        if call_fn is not None:
            reloaded_campaign.call_for_each_run(call_fn)
        if actions is not None:
            reloaded_campaign.apply_for_each_run_dir(actions)
        logging.debug("Completed runs:")
        logging.debug(pformat(reloaded_campaign.scan_completed()))
        logging.debug("All completed? %s", str(reloaded_campaign.all_complete()))
        reloaded_campaign.collate()
        logging.debug("data:\n %s", str(reloaded_campaign.get_collation_result()))
        logging.debug(reloaded_campaign)
        # Create a BasicStats analysis element and apply it to the campaign
        if stats is not None:
            reloaded_campaign.apply_analysis(stats)
            logging.debug("stats:\n %s", str(reloaded_campaign.get_last_analysis()))
        # Print the campaign log
        logging.debug(pformat(reloaded_campaign._log))
        logging.debug("All completed? %s", str(reloaded_campaign.all_complete()))
    return _campaign


def test_cannonsim(tmpdir, campaign):
    # Define parameter space for the cannonsim app
    params = {
        "angle": {
            "type": "float",
            "min": 0.0,
            "max": 6.28,
            "default": 0.79},
        "air_resistance": {
            "type": "float",
            "min": 0.0,
            "max": 1.0,
            "default": 0.2},
        "height": {
            "type": "float",
            "min": 0.0,
            "max": 1000.0,
            "default": 1.0},
        "time_step": {
            "type": "float",
            "min": 0.0001,
            "max": 1.0,
            "default": 0.01},
        "gravity": {
            "type": "float",
            "min": 0.0,
            "max": 1000.0,
            "default": 9.8},
        "mass": {
            "type": "float",
            "min": 0.0001,
            "max": 1000.0,
            "default": 1.0},
        "velocity": {
            "type": "float",
            "min": 0.0,
            "max": 1000.0,
            "default": 10.0}}

    # Create an encoder and decoder for the cannonsim app
    encoder = uq.encoders.GenericEncoder(
        template_fname='tests/cannonsim/test_input/cannonsim.template',
        delimiter='#',
        target_filename='in.cannon')
    decoder = uq.decoders.SimpleCSV(
        target_filename='output.csv', output_columns=[
            'Dist', 'lastvx', 'lastvy'], header=0)
    # Create a collation element for this campaign
    collater = uq.collate.AggregateSamples(average=False)
    actions = uq.actions.ExecuteLocal("tests/cannonsim/bin/cannonsim in.cannon output.csv")
    stats = uq.analysis.BasicStats(qoi_cols=['Dist', 'lastvx', 'lastvy'])
    # Make a random sampler
    vary = {
        "angle": cp.Uniform(0.0, 1.0),
        "height": cp.Uniform(2.0, 10.0),
        "velocity": cp.Normal(10.0, 1.0),
        "mass": cp.Uniform(5.0, 1.0)
    }
    sampler = uq.sampling.RandomSampler(vary=vary)
    campaign(tmpdir, 'cannon', 'cannonsim', params, encoder, decoder, sampler,
             collater, actions, stats, vary, 5, 1)
    # campaign(tmpdir, 'cannon', 'cannonsim', params, encoder, decoder, sampler,
    #         collater, None, stats, vary, 5, 1, call_fn=execute_cannonsim)
    # Make a sweep sampler
    sweep = {
        "angle": [0.1, 0.2, 0.3],
        "height": [2.0, 10.0],
        "velocity": [10.0, 10.1, 10.2]
    }
    sampler = uq.sampling.BasicSweep(sweep=sweep)
    campaign(tmpdir, 'cannonsim', 'cannonsim', params, encoder, decoder, sampler,
             collater, actions, None, sweep, 5, 1)


def test_gauss(tmpdir, campaign):
    params = {
        "sigma": {
            "type": "float",
            "min": 0.0,
            "max": 100000.0,
            "default": 0.25
        },
        "mu": {
            "type": "float",
            "min": 0.0,
            "max": 100000.0,
            "default": 1
        },
        "num_steps": {
            "type": "integer",
            "min": 0,
            "max": 100000,
            "default": 10
        },
        "out_file": {
            "type": "string",
            "default": "output.csv"
        },
        "bias": {
            "type": "fixture",
            "allowed": ["bias1", "bias2"],
            "default": "bias1"
        }
    }
    fixtures = {
        "bias1": {
            "type": "file", "path": "tests/gauss/bias1.txt",
            "common": False, "exists_local": True,
            "target": "",
            "group": ""
        },
        "bias2": {
            "type": "file", "path": "tests/gauss/bias2.txt",
            "common": False, "exists_local": True,
            "target": "",
            "group": ""
        }
    }
    encoder = uq.encoders.GenericEncoder(template_fname='tests/gauss/gauss.template',
                                         target_filename='gauss_in.json')
    fixtures_encoder = uq.encoders.ApplyFixtures(fixtures=fixtures)
    encoder_with_fixtures = uq.encoders.MultiEncoder(encoder, fixtures_encoder)

    decoder = GaussDecoder(target_filename=params['out_file']['default'])
    collater = uq.collate.AggregateSamples(average=False)
    actions = uq.actions.ExecuteLocal("tests/gauss/gauss_json.py gauss_in.json")
    stats = uq.analysis.EnsembleBoot(groupby=["mu"], qoi_cols=["numbers"])
    vary = {
        "mu": cp.Uniform(1.0, 100.0),
    }
    sampler = uq.sampling.RandomSampler(vary=vary)
    campaign(tmpdir, 'gauss', 'gauss', params, encoder, decoder, sampler,
             collater, actions, stats, vary, 2, 2)
    encoder = GaussEncoder(target_filename='gauss_in.json')
    campaign(tmpdir, 'gauss', 'gauss', params, encoder, decoder, sampler,
             collater, actions, stats, vary, 2, 2)
    campaign(tmpdir, 'gauss', 'gauss', params, encoder_with_fixtures, decoder, sampler,
             collater, actions, stats, vary, 2, 2)


def test_pce(tmpdir, campaign):
    # Define parameter space
    params = {
        "temp_init": {
            "type": "float",
            "min": 0.0,
            "max": 100.0,
            "default": 95.0},
        "kappa": {
            "type": "float",
            "min": 0.0,
            "max": 0.1,
            "default": 0.025},
        "t_env": {
            "type": "float",
            "min": 0.0,
            "max": 40.0,
            "default": 15.0},
        "out_file": {
            "type": "string",
            "default": "output.csv"}}
    output_filename = params["out_file"]["default"]
    output_columns = ["te"]
    # Create an encoder and decoder for PCE test app
    encoder = uq.encoders.GenericEncoder(
        template_fname='tests/cooling/cooling.template',
        delimiter='$',
        target_filename='cooling_in.json')
    decoder = uq.decoders.SimpleCSV(target_filename=output_filename,
                                    output_columns=output_columns,
                                    header=0)
    # Create a collation element for this campaign
    collater = uq.collate.AggregateSamples(average=False)
    # Create the sampler
    vary = {
        "kappa": cp.Uniform(0.025, 0.075),
        "t_env": cp.Uniform(15, 25)
    }
    sampler = uq.sampling.PCESampler(vary=vary,
                                     polynomial_order=3)
    actions = uq.actions.ExecuteLocal("tests/cooling/cooling_model.py cooling_in.json")
    stats = uq.analysis.PCEAnalysis(sampler=sampler,
                                    qoi_cols=output_columns)
    campaign(tmpdir, 'pce', 'pce', params, encoder, decoder, sampler,
             collater, actions, stats, vary, 0, 1)


def test_sc(tmpdir, campaign):
    params = {
        "Pe": {
            "type": "float",
            "min": "1.0",
            "max": "2000.0",
            "default": "100.0"},
        "f": {
            "type": "float",
            "min": "0.0",
            "max": "10.0",
            "default": "1.0"},
        "out_file": {
            "type": "string",
            "default": "output.csv"}}
    output_filename = params["out_file"]["default"]
    output_columns = ["u"]
    encoder = uq.encoders.GenericEncoder(
        template_fname=f'tests/sc/sc.template',
        delimiter='$',
        target_filename='sc_in.json')
    decoder = uq.decoders.SimpleCSV(target_filename=output_filename,
                                    output_columns=output_columns,
                                    header=0)
    collater = uq.collate.AggregateSamples(average=False)
    vary = {
        "Pe": cp.Uniform(100.0, 200.0),
        "f": cp.Normal(1.0, 0.1)
    }
    sampler = uq.sampling.SCSampler(vary=vary, polynomial_order=1)
    actions = uq.actions.ExecuteLocal(f"tests/sc/sc_model.py sc_in.json")
    stats = uq.analysis.SCAnalysis(sampler=sampler, qoi_cols=output_columns)
    campaign(tmpdir, 'sc', 'sc', params, encoder, decoder, sampler,
             collater, actions, stats, vary, 0, 1)


# def test_qmc(tmpdir, campaign):
#     # Define parameter space
#     params = {
#         "temp_init": {
#             "type": "float",
#             "min": 0.0,
#             "max": 100.0,
#             "default": 95.0},
#         "kappa": {
#             "type": "float",
#             "min": 0.0,
#             "max": 0.1,
#             "default": 0.025},
#         "t_env": {
#             "type": "float",
#             "min": 0.0,
#             "max": 40.0,
#             "default": 15.0},
#         "out_file": {
#             "type": "string",
#             "default": "output.csv"}}
#     output_filename = params["out_file"]["default"]
#     output_columns = ["te"]
#     # Create an encoder and decoder for QMC test app
#     encoder = uq.encoders.GenericEncoder(
#         template_fname='tests/cooling/cooling.template',
#         delimiter='$',
#         target_filename='cooling_in.json')
#     decoder = uq.decoders.SimpleCSV(target_filename=output_filename,
#                                     output_columns=output_columns,
#                                     header=0)
#     # Create a collation element for this campaign
#     collater = uq.collate.AggregateSamples(average=False)
#     # Create the sampler
#     vary = {
#         "kappa": cp.Uniform(0.025, 0.075),
#         "t_env": cp.Uniform(15, 25)
#     }
#     sampler = uq.sampling.QMCSampler(vary=vary,
#                                      n_mc_samples=10)
#     actions = uq.actions.ExecuteLocal("tests/cooling/cooling_model.py cooling_in.json")
#     stats = uq.analysis.QMCAnalysis(sampler=sampler,
#                                     qoi_cols=output_columns)
#     campaign(tmpdir, 'qmc2', 'qmc2', params, encoder, decoder, sampler,
#              collater, actions, stats, vary, 10, 1)



import easyvvuq as uq
import chaospy as cp
import os
import sys
import pytest
from easyvvuq.encoders.jinja_encoder import JinjaEncoder

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


def test_jinjaencoder(tmpdir):
    """
    Set up a campaign using the jinja2 template.
    This example is based on the DALES model. The input file is a Fortran namelist.
    """

    params = {
        "Nc_0": {  # concentration of cloud condensation nuclei
            "type": "float",
            "min": 0.1e6,
            "max": 1000e6,
            "default": 70e6,
        },
        "cf": {  # cf subgrid filter constant
            "type": "float",
            "min": 1.0,
            "max": 4.0,
            "default": 2.5,
        },
        "cn": {  # subfilterscale parameter
            "type": "float",
            "min": 0.5,
            "max": 1.0,
            "default": 0.76,
        },
        "Rigc": {  # critical Richardson number
            "type": "float",
            "min": 0.1,
            "max": 1.0,
            "default": 0.25,
        },
        "Prandtl": {  # Prandtl number, subgrid.
            "type": "float",
            "min": 0.1,
            "max": 1.0,
            "default": 1.0 / 3,
        },
        "z0": {  # surface roughness
            "type": "float",
            "min": 1e-4,
            "max": 1.0,
            "default": 1.6e-4,
        },
        "l_sb": {  # flag for microphysics scheme: false - KK00 Khairoutdinov and Kogan, 2000
            "type": "integer",    # true - SB   Seifert and Beheng, 2001, 2006, Default
            "min": 0,
            "max": 1,
            "default": 1
        },
        "Nh": {  # number of grid points in the horizontal directions - itot, jtot
            "type": "integer",
            "min": 3,
            "max": 1024,
            "default": 10
        },
        "extent": {  # norizontal domain size in x, y  - xsize, ysize. unit: m
            "type": "float",
            "min": 1,
            "max": 1000000,
            "default": 1000,
        },
        "seed": {  # random seed
            "type": "integer",
            "min": 1,
            "max": 1000000,
            "default": 43
        },
        "nprocx": {  # number of MPI tasks in x
            "type": "integer",
            "min": 1,
            "max": 1000,
            "default": 1
        },
        "nprocy": {  # number of MPI tasks in y
            "type": "integer",
            "min": 1,
            "max": 1000,
            "default": 1
        },
    }

    vary = {
        "Nc_0": cp.Uniform(50e6, 100e6),
        # "cf": cp.Uniform(2.4, 2.6),
        # "cn": cp.Uniform(0.5, 0.9),
        # "Rigc": cp.Uniform(0.1, 0.4),
        # "Prandtl": cp.Uniform(0.2, 0.4),
        # "z0": cp.Uniform(1e-4, 2e-4),
        "l_sb": cp.DiscreteUniform(0, 1),
        # "Nh": cp.DiscreteUniform(10, 20),
        # "extent": cp.Uniform(1000, 2000),
        "seed": cp.Uniform(1, 2000),
    }

    output_columns = ['cfrac', 'lwp', 'rwp', 'zb', 'zi', 'prec', 'wq', 'wtheta', 'we', 'walltime']
    my_sampler = uq.sampling.SCSampler(vary=vary, polynomial_order=2,
                                       quadrature_rule="C")
    my_campaign = uq.Campaign(name='dales', work_dir=tmpdir, db_location='sqlite:///')
    encoder = JinjaEncoder(template_fname='tests/jinjaencoder/namoptions.template',
                           target_filename='namoptions.001')
    decoder = uq.decoders.SimpleCSV(
        target_filename='results.csv',
        output_columns=output_columns,
        header=0)
    collater = uq.collate.AggregateSamples(average=False)
    my_campaign.add_app(name="dales",
                        params=params,
                        encoder=encoder,
                        decoder=decoder,
                        collater=collater)
    my_campaign.verify_all_runs = False  # to prevent errors on integer quantities
    my_campaign.set_sampler(my_sampler)
    my_campaign.draw_samples()
    my_campaign.populate_runs_dir()


if __name__ == "__main__":
    test_jinjaencoder("/tmp/")



"""Provides class that allows access to an SQL format CampaignDB.
"""
import os
import json
import logging
import pandas as pd
from sqlalchemy import create_engine, Column, Integer, String, ForeignKey
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy import MetaData
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import text
from .base import BaseCampaignDB
from easyvvuq import constants
from easyvvuq.sampling.base import BaseSamplingElement
from easyvvuq.encoders.base import BaseEncoder
from easyvvuq.decoders.base import BaseDecoder
from easyvvuq.collate.base import BaseCollationElement
from easyvvuq import ParamsSpecification
from easyvvuq.utils.helpers import multi_index_tuple_parser

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"

logger = logging.getLogger(__name__)

Base = declarative_base()


class DBInfoTable(Base):
    """An SQLAlchemy schema for the database information table.
    """
    __tablename__ = 'db_info'
    id = Column(Integer, primary_key=True)
    next_run = Column(Integer)
    next_ensemble = Column(Integer)


class CampaignTable(Base):
    """An SQLAlchemy schema for the campaign information table.
    """
    __tablename__ = 'campaign_info'
    id = Column(Integer, primary_key=True)
    name = Column(String)
    easyvvuq_version = Column(String)
    campaign_dir_prefix = Column(String)
    campaign_dir = Column(String)
    runs_dir = Column(String)
    sampler = Column(Integer, ForeignKey('sampler.id'))


class AppTable(Base):
    """An SQLAlchemy schema for the app table.
    """
    __tablename__ = 'app'
    id = Column(Integer, primary_key=True)
    name = Column(String)
    input_encoder = Column(String)
    output_decoder = Column(String)
    collater = Column(String)
    params = Column(String)


class RunTable(Base):
    """An SQLAlchemy schema for the run table.
    """
    __tablename__ = 'run'
    id = Column(Integer, primary_key=True)
    run_name = Column(String)
    ensemble_name = Column(String)
    app = Column(Integer, ForeignKey('app.id'))
    params = Column(String)
    status = Column(Integer)
    run_dir = Column(String)
    campaign = Column(Integer, ForeignKey('campaign_info.id'))
    sampler = Column(Integer, ForeignKey('sampler.id'))


class SamplerTable(Base):
    """An SQLAlchemy schema for the run table.
    """
    __tablename__ = 'sampler'
    id = Column(Integer, primary_key=True)
    sampler = Column(String)


class CampaignDB(BaseCampaignDB):

    def __init__(self, location=None, new_campaign=False, name=None, info=None):

        if location is not None:
            self.engine = create_engine(location)
        else:
            self.engine = create_engine('sqlite://')

        session_maker = sessionmaker(bind=self.engine)

        self.session = session_maker()

        if new_campaign:
            if info is None:
                raise RuntimeError('No information provided to create'
                                   'database')
            if info.name != name:
                message = (f'Information for campaign {info.name} given '
                           f'for campaign database {name}')
                logging.critical(message)
                raise RuntimeError(message)

            Base.metadata.create_all(self.engine)

            is_db_empty = (self.session.query(CampaignTable).first() is None)

            version_check = self.session.query(
                CampaignTable).filter(CampaignTable.easyvvuq_version != info.easyvvuq_version).all()

            if (not is_db_empty) and (len(version_check) != 0):
                raise RuntimeError('Database contains campaign created with an incompatible' +
                                   ' version of EasyVVUQ!')

            self._next_run = 1
            self._next_ensemble = 1

            self.session.add(CampaignTable(**info.to_dict(flatten=True)))
            self.session.add(
                DBInfoTable(
                    next_run=self._next_run,
                    next_ensemble=self._next_ensemble))
            self.session.commit()
        else:
            info = self.session.query(
                CampaignTable).filter_by(name=name).first()
            if info is None:
                raise ValueError('Campaign with the given name not found.')

            db_info = self.session.query(DBInfoTable).first()
            self._next_run = db_info.next_run
            self._next_ensemble = db_info.next_ensemble

    def app(self, name=None):
        """
        Get app information. Specific applications selected by `name`,
        otherwise first entry in database 'app' selected.

        Parameters
        ----------
        name : str or None
            Name of selected app, if `None` given then first app will be
            selected.

        Returns
        -------
        dict:
            Application information.
        """

        if name is None:
            logging.warning('No app name provided so using first app '
                            'in database')
            selected = self.session.query(AppTable).all()
        else:
            selected = self.session.query(AppTable).filter_by(name=name).all()

        if len(selected) == 0:
            message = f'No entry for app: ({name}).'
            logger.critical(message)
            raise RuntimeError(message)
        if len(selected) > 1:
            message = f'Too many apps called: ({name}).'
            logger.critical(message)
            raise RuntimeError(message)

        selected_app = selected[0]

        app_dict = {
            'id': selected_app.id,
            'name': selected_app.name,
            'input_encoder': selected_app.input_encoder,
            'output_decoder': selected_app.output_decoder,
            'collater': selected_app.collater,
            'params': ParamsSpecification.deserialize(selected_app.params)
        }

        return app_dict

    def add_app(self, app_info):
        """
        Add application to the 'app' table.

        Parameters
        ----------
        app_info: AppInfo
            Application definition.

        Returns
        -------

        """

        # Check that no app with same name exists
        name = app_info.name
        selected = self.session.query(AppTable).filter_by(name=name).all()
        if len(selected) > 0:
            message = (
                f'There is already an app in this database with name {name}'
                f'(found {len(selected)}).'
            )
            logger.critical(message)
            raise RuntimeError(message)

        app_dict = app_info.to_dict(flatten=True)

        db_entry = AppTable(**app_dict)
        self.session.add(db_entry)
        self.session.commit()

    def add_sampler(self, sampler_element):
        """
        Add new Sampler to the 'sampler' table.

        Parameters
        ----------
        sampler_element: BaseSamplingElement

        Returns
        -------

        """
        db_entry = SamplerTable(sampler=sampler_element.serialize())

        self.session.add(db_entry)
        self.session.commit()

        return db_entry.id

    def update_sampler(self, sampler_id, sampler_element):
        """
        Update the state of the Sampler with id 'sampler_id' to
        that in the passed 'sampler_element'

        Parameters
        ----------
        sampler_id: int
            The id of the sampler in the db to update
        sampler_element: BaseSamplingElement
            The sampler whose state should be used as the new state

        Returns
        -------

        """

        selected = self.session.query(SamplerTable).get(sampler_id)
        selected.sampler = sampler_element.serialize()
        self.session.commit()

    def resurrect_sampler(self, sampler_id):
        """
        Return the sampler object corresponding to id sampler_id in the database.
        It is deserialized from the state stored in the database.

        Parameters
        ----------
        sampler_id: int
            The id of the sampler to resurrect

        Returns
        -------
        BaseSamplingElement
            The 'live' sampler object, deserialized from the state in the db

        """

        serialized_sampler = self.session.query(SamplerTable).get(sampler_id).sampler
        sampler = BaseSamplingElement.deserialize(serialized_sampler)
        return sampler

    def resurrect_app(self, app_name):
        """
        Return the 'live' encoder, decoder and collation objects corresponding to the app with
        name 'app_name' in the database. They are deserialized from the states previously
        stored in the database.

        Parameters
        ----------
        app_name: string
            Name of the app to resurrect

        Returns
        -------
        BaseEncoder, BaseDecoder, BaseCollationElement
            The 'live' encoder and decoder objects associated with this app

        """

        app_info = self.app(app_name)

        encoder = BaseEncoder.deserialize(app_info['input_encoder'])
        decoder = BaseDecoder.deserialize(app_info['output_decoder'])
        collater = BaseCollationElement.deserialize(app_info['collater'])
        return encoder, decoder, collater

    def add_runs(self, run_info_list=None, run_prefix='Run_', ensemble_prefix='Ensemble_'):
        """
        Add list of runs to the `runs` table in the database.

        Parameters
        ----------
        run_info_list: List of RunInfo objects
            Each RunInfo object contains relevant run fields: params, status (where in the
            EasyVVUQ workflow is this RunTable), campaign (id number), sample, app
        run_prefix: str
            Prefix for run id
        ensemble_prefix: str
            Prefix for ensemble id

        Returns
        -------

        """

        # Add all runs to RunTable
        runs_dir = self.runs_dir()
        for run_info in run_info_list:
            run_info.ensemble_name = f"{ensemble_prefix}{self._next_ensemble}"
            run_info.run_name = f"{run_prefix}{self._next_run}"
            run_info.run_dir = os.path.join(runs_dir, run_info.run_name)

            run = RunTable(**run_info.to_dict(flatten=True))
            self.session.add(run)
            self._next_run += 1
        self._next_ensemble += 1

        # Update run and ensemble counters in db
        db_info = self.session.query(DBInfoTable).first()
        db_info.next_run = self._next_run
        db_info.next_ensemble = self._next_ensemble

        self.session.commit()

    @staticmethod
    def _run_to_dict(run_row):
        """
        Convert the provided row from 'runs' table into a dictionary

        Parameters
        ----------
        run_row: RunTable
            Information on a particular run in the database.

        Returns
        -------
        dict:
            Contains run information (keys = run_name, params, status, sample,
            campaign and app)

        """

        run_info = {
            'run_name': run_row.run_name,
            'ensemble_name': run_row.ensemble_name,
            'params': json.loads(run_row.params),
            'status': constants.Status(run_row.status),
            'sampler': run_row.sampler,
            'campaign': run_row.campaign,
            'app': run_row.app,
            'run_dir': run_row.run_dir
        }

        return run_info

    def set_dir_for_run(self, run_name, run_dir, campaign=None, sampler=None):
        """
        Set the 'run_dir' path for the specified run in the database.

        Parameters
        ----------
        run_name: str
            Name of run to filter for.
        run_dir: str
            Directory path associated to set for this run.
        campaign:  int or None
            Campaign id to filter for.
        sampler: int or None
            Sample id to filter for.

        Returns
        -------

        """

        filter_options = {'run_name': run_name}
        if campaign:
            filter_options['campaign'] = campaign
        if sampler:
            filter_options['sampler'] = sampler

        selected = self.session.query(RunTable).filter_by(**filter_options)

        if selected.count() != 1:
            logging.critical('Multiple runs selected - using the first')

        selected = selected.first()

        selected.run_dir = run_dir
        self.session.commit()

    def get_run_status(self, run_name, campaign=None, sampler=None):
        """
        Return the status (enum) for the run with name 'run_name' (and, optionally,
        filtering for campaign and sampler by id)

        Parameters
        ----------
        run_name: str
            Name of the run
        campaign: int
            ID of the desired Campaign
        sampler: int
            ID of the desired Sampler

        Returns
        -------
        status: enum(Status)
            Status of the run.
        """

        filter_options = {'run_name': run_name}
        if campaign:
            filter_options['campaign'] = campaign
        if sampler:
            filter_options['sampler'] = sampler

        selected = self.session.query(RunTable).filter_by(**filter_options)

        if selected.count() != 1:
            logging.critical('Multiple runs selected - using the first')

        selected = selected.first()

        return constants.Status(selected.status)

    def set_run_statuses(self, run_name_list, status):
        """
        Set the specified 'status' (enum) for all runs in the list run_ID_list

        Parameters
        ----------
        run_name_list: list of str
            A list of run names run names (format is usually: prefix + int)
        status: enum(Status)
            The new status all listed runs should now have

        Returns
        -------

        """
        max_entries = 900

        for i in range(0, len(run_name_list), max_entries):
            selected = self.session.query(RunTable).filter(
                RunTable.run_name.in_(set(run_name_list[i:i + max_entries]))).all()

            for run in selected:
                run.status = status
            self.session.commit()

    def campaigns(self):
        """Get list of campaigns for which information is stored in the
        database.

        Returns
        -------
        list:
            Campaign names.
        """

        return [c.name for c in self.session.query(CampaignTable).all()]

    def _get_campaign_info(self, campaign_name=None):
        """
        Parameters
        ----------
        campaign_name: str
            Name of campaign to select

        Returns
        -------
            sqlalchemy query for campaign with this name

        """

        query = self.session.query(CampaignTable)

        if campaign_name is None:
            campaign_info = query
        else:
            campaign_info = query.filter_by(name=campaign_name).all()

        if campaign_name is not None:
            if len(campaign_info) > 1:
                logger.warning(
                    'More than one campaign selected - using first one.')
            elif len(campaign_info) == 0:
                message = 'No campaign available.'
                logger.critical(message)
                raise RuntimeError(message)
            return campaign_info[0]

        return campaign_info.first()

    def get_campaign_id(self, name):
        """
        Return the (database) id corresponding to the campaign with name 'name'.

        Parameters
        ----------
        name: str
            Name of the campaign.

        Returns
        -------
        int:
            The id of the campaign with the specified name
        """

        selected = self.session.query(
            CampaignTable.name.label(name),
            CampaignTable.id).filter(CampaignTable.name == name).all()
        if len(selected) == 0:
            msg = f"No campaign with name {name} found in campaign database"
            logger.error(msg)
            raise RuntimeError(msg)
        if len(selected) > 1:
            msg = (
                f"More than one campaign with name {name} found in"
                f"campaign database. Database state is compromised."
            )
            logger.error(msg)
            raise RuntimeError(msg)

        # Return the database ID for the specified campaign
        return selected[0][1]

    def get_sampler_id(self, campaign_id):
        """
        Return the (database) id corresponding to the sampler currently set
        for the campaign with id 'campaign_id'

        Parameters
        ----------
        campaign_id: int
            ID of the campaign.

        Returns
        -------
        int:
            The id of the sampler set for the specified campaign
        """

        sampler_id = self.session.query(CampaignTable).get(campaign_id).sampler
        return sampler_id

    def set_sampler(self, campaign_id, sampler_id):
        """
        Set specified campaign to be using specified sampler

        Parameters
        ----------
        campaign_id: int
            ID of the campaign.
        sampler_id: int
            ID of the sampler.

        Returns
        -------
        """

        self.session.query(CampaignTable).get(campaign_id).sampler = sampler_id
        self.session.commit()

    def campaign_dir(self, campaign_name=None):
        """Get campaign directory for `campaign_name`.

        Parameters
        ----------
        campaign_name: str
            Name of campaign to select

        Returns
        -------
        str:
            Path to campaign directory.
        """

        return self._get_campaign_info(campaign_name=campaign_name).campaign_dir

    def _select_runs(
            self,
            name=None,
            campaign=None,
            sampler=None,
            status=None,
            not_status=None,
            app_id=None):
        """
        Select all runs in the database which match the input criteria.

        Parameters
        ----------
        name: str
            Name of run to filter for.
        campaign:  int or None
            Campaign id to filter for.
        sampler: int or None
            Sampler id to filter for.
        status: enum(Status) or None
            Status string to filter for.
        not_status: enum(Status) or None
            Exclude runs with this status string

        Returns
        -------
        sqlalchemy.orm.query.Query
            Selected runs from the database run table.
        """
        filter_options = {}
        if name:
            filter_options['run_name'] = name
        if campaign:
            filter_options['campaign'] = campaign
        if sampler:
            filter_options['sampler'] = sampler
        if status:
            filter_options['status'] = status
        if app_id:
            filter_options['app'] = app_id

        # Note that for some databases this can be sped up with a yield_per(), but not all
        selected = self.session.query(RunTable).filter_by(
            **filter_options).filter(RunTable.status != not_status)

        return selected

    def run(self, name, campaign=None, sampler=None, status=None, not_status=None, app_id=None):
        """
        Get the information for a specified run.

        Parameters
        ----------
        name: str
            Name of run to filter for.
        campaign:  int or None
            Campaign id to filter for.
        sampler: int or None
            Sampler id to filter for.
        status: enum(Status) or None
            Status string to filter for.
        not_status: enum(Status) or None
            Exclude runs with this status string

        Returns
        -------
        dict
            Containing run information (run_name, params, status, sample,
            campaign, app)
        """

        selected = self._select_runs(
            name=name,
            campaign=campaign,
            sampler=sampler,
            status=status,
            not_status=not_status,
            app_id=app_id)

        if selected.count() != 1:
            logging.warning('Multiple runs selected - using the first')

        selected = selected.first()

        return self._run_to_dict(selected)

    def runs(self, campaign=None, sampler=None, status=None, not_status=None, app_id=None):
        """
        A generator to return all run information for selected `campaign` and `sampler`.

        Parameters
        ----------
        campaign: int or None
            Campaign id to filter for.
        sampler: int or None
            Sampler id to filter for.
        status: enum(Status) or None
            Status string to filter for.
        not_status: enum(Status) or None
            Exclude runs with this status string

        Returns
        -------
        dict:
            Information on each selected run (key = run_name, value = dict of
            run information fields.), one at a time.

        """

        selected = self._select_runs(
            campaign=campaign,
            sampler=sampler,
            status=status,
            not_status=not_status,
            app_id=app_id)

        for r in selected:
            yield r.run_name, self._run_to_dict(r)

    def run_ids(self, campaign=None, sampler=None, status=None, not_status=None, app_id=None):
        """
        A generator to return all run IDs for selected `campaign` and `sampler`.

        Parameters
        ----------
        campaign: int or None
            Campaign id to filter for.
        sampler: int or None
            Sampler id to filter for.
        status: enum(Status) or None
            Status string to filter for.
        not_status: enum(Status) or None
            Exclude runs with this status string

        Returns
        -------
        str:
            run ID for each selected run, one at a time.

        """

        selected = self._select_runs(
            campaign=campaign,
            sampler=sampler,
            status=status,
            not_status=not_status,
            app_id=app_id)

        for r in selected:
            yield r.run_name

    def get_num_runs(self, campaign=None, sampler=None, status=None, not_status=None):
        """
        Returns the number of runs matching the filtering criteria.

        Parameters
        ----------
        campaign: int or None
            Campaign id to filter for.
        sampler: int or None
            Sampler id to filter for.
        status: enum(Status) or None
            Status string to filter for.
        not_status: enum(Status) or None
            Exclude runs with this status string

        Returns
        -------
        int:
            The number of runs in the database matching the filtering criteria

        """

        selected = self._select_runs(
            campaign=campaign,
            sampler=sampler,
            status=status,
            not_status=not_status)

        return selected.count()

    def runs_dir(self, campaign_name=None):
        """
        Get the directory used to store run information for `campaign_name`.

        Parameters
        ----------
        campaign_name: str
            Name of the selected campaign.

        Returns
        -------
        str:
            Path containing run outputs.
        """

        return self._get_campaign_info(campaign_name=campaign_name).runs_dir

    def append_collation_dataframe(self, df, app_id):
        """
        Append the data in dataframe 'df' to that already collated in the database
        for the specified app.

        Parameters
        ----------
        df: pandas dataframe
            The dataframe whose contents need to be appended to the collation store
        app_id: int
            The id of the app in the sql database. Used to determine which collation
            table is appended to.

        Returns
        -------
        """

        if df.size == 0:
            logging.warning(
                f"Attempt to append empty dataframe to SQL collation table for app_id {app_id}.")
            return

        tablename = 'COLLATION_APP' + str(app_id)
        df.to_sql(tablename, self.engine, if_exists='append')

    def get_collation_dataframe(self, app_id):
        """
        Returns a dataframe containing the full collated results stored in this database
        for the specified app.
        i.e. the total of what was added with the append_collation_dataframe() method.

        Parameters
        ----------
        app_id: int
            The id of the app in the sql database. Used to determine which collation
            table is appended to.

        Returns
        -------
        df: pandas dataframe
            The dataframe with all contents that were appended to this database
        """

        tablename = 'COLLATION_APP' + str(app_id)
        if tablename in self.engine.table_names():
            query = "select * from " + tablename
            df = pd.read_sql_query(query, self.engine.execution_options(sqlite_raw_colnames=True))
            columns, multi = multi_index_tuple_parser(df.columns.values[1:])
            if multi:
                df = pd.DataFrame(df.values[:, 1:],
                                  columns=pd.MultiIndex.from_tuples(columns))
            return df
        else:
            return None

    def clear_collation(self, app_id):
        tablename = 'COLLATION_APP' + str(app_id)

        if tablename in self.engine.table_names():
            sqlcmd = text(f'DROP TABLE {tablename};')
            self.engine.execute(sqlcmd)



"""Data structures to ensure consistency during serialization for databases.

"""
import logging
import cerberus
import json
import numpy

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"

logger = logging.getLogger(__name__)


class EasyVVUQValidator(cerberus.Validator):
    def __init__(self, *args, **kwargs):
        super(EasyVVUQValidator, self).__init__(*args, **kwargs)

        # Add numpy.int64 as an acceptable 'integer' type
        integer_type = cerberus.TypeDefinition('integer', (int, numpy.int64), ())
        cerberus.Validator.types_mapping['integer'] = integer_type

        # Add 'fixture' type (for now, it's expected just to be a string)
        fixture_type = cerberus.TypeDefinition('fixture', (str), ())
        cerberus.Validator.types_mapping['fixture'] = fixture_type


class ParamsSpecification:

    def __init__(self, params, appname=None):

        if not isinstance(params, dict):
            msg = "params must be of type 'dict'"
            logger.error(msg)
            raise Exception(msg)

        if not params:
            msg = ("params must not be empty. At least one parameter "
                   "should be specified.")
            logger.error(msg)
            raise Exception(msg)

        # Check each param has a dict as a value, and that dict has a "default" defined
        for param_key, param_def in params.items():
            if not isinstance(param_def, dict):
                msg = f"Entry for param '{param_key}' must be a dictionary"
                logger.error(msg)
                raise Exception(msg)
            if "default" not in param_def:
                msg = (
                    f"Entry for param '{param_key}' must be a dictionary"
                    f"defining a 'default' value for this parameter."
                )
                logger.error(msg)
                raise Exception(msg)

        self.params_dict = params
        self.appname = appname

        # Create a validator for the schema defined by params_dict
        self.cerberus_validator = EasyVVUQValidator(self.params_dict)

    def process_run(self, new_run, verify=True):

        # If necessary parameter names are missing, fill them in from the
        # default values in params_info
        for param in self.params_dict.keys():
            if param not in new_run.keys():
                default_val = self.params_dict[param]["default"]
                new_run[param] = default_val

        # Optionally verify that all params are known for this app, that the types are
        # correct, params are within specified ranges etc. Uses cerberus for this.
        if verify:
            if not self.cerberus_validator.validate(new_run):
                errors = self.cerberus_validator.errors
                msg = (
                    f"Error when verifying the following new run:\n"
                    f"{new_run}\n"
                    f"Identified errors were:\n"
                    f"{errors}\n")

                errors_list = [error[0] for error in self.cerberus_validator.errors.values()]
                if 'unknown field' in errors_list:
                    msg += (
                        f"The allowed parameter names for this app are:\n"
                        f"{list(self.params_dict.keys())}")

                logger.error(msg)
                raise RuntimeError(msg)

        return new_run

    def serialize(self):
        return json.dumps(self.params_dict)

    @staticmethod
    def deserialize(serialized_params):
        return ParamsSpecification(json.loads(serialized_params))



from .base import BaseSamplingElement
import logging
import itertools
import functools

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"

logger = logging.getLogger(__name__)


class MultiSampler(BaseSamplingElement, sampler_name="multisampler"):

    def __init__(self, *samplers, count=0, serialized_list_of_samplers=None):
        """
            Expects one or more samplers
        """

        # If no serialized samplers list passed, generate one. Else deserialize the passed samplers.
        if len(samplers) < 1 and serialized_list_of_samplers is None:
            raise RuntimeError("You need to supply at least one sampler to the MultiSampler")
        if serialized_list_of_samplers is None:
            self.samplers = samplers
            self.serialized_list_of_samplers = [sampler.serialize() for sampler in self.samplers]
        else:
            self.serialized_list_of_samplers = serialized_list_of_samplers
            self.samplers = []
            for serialized_sampler in self.serialized_list_of_samplers:
                self.samplers.append(BaseSamplingElement.deserialize(serialized_sampler))

        # Multisampler is finite only if all samplers in it are finite
        for sampler in self.samplers:
            if sampler.is_finite() is False:
                msg = "Multisampler must be composed of finite samplers"
                logger.critical(msg)
                raise RuntimeError(msg)

        # Combine all the iterables/generators into one
        self.multi_iterator = itertools.product(*self.samplers)

        self.count = 0
        for i in range(count):
            try:
                self.__next__()
            except StopIteration:
                logger.warning("Multisampler constructed, but has no samples left to draw.")

    def element_version(self):
        return "0.1"

    def is_finite(self):
        return True

    def n_samples(self):
        """Returns the number of samples in this sampler.

        Returns
        -------
        a product of the sizes of samplers passed to MultiSampler
        """
        return functools.reduce(
            lambda x, y: x * y, [sampler.n_samples() for sampler in self.samplers], 1)

    def __next__(self):
        # Will raise StopIteration when there are none left
        multisampler_run = self.multi_iterator.__next__()

        run_dict = {}
        for contribution in multisampler_run:
            run_dict = {**run_dict, **contribution}

        self.count += 1
        return run_dict

    def is_restartable(self):
        return True

    def get_restart_dict(self):
        return {'serialized_list_of_samplers': self.serialized_list_of_samplers}



from easyvvuq import OutputType
from .base import BaseAnalysisElement
from sklearn.gaussian_process import GaussianProcessRegressor


class GaussianProcessSurrogate(BaseAnalysisElement):

    def __init__(self, attr_cols, target_cols):
        """Element to calculate basic stats for `qoi_cols` values.

        This results in values for: count, mean, std, min, max and 25%, 50% &
        75% percentiles for each value in the analysis.

        Parameters
        ----------
        attr_cols : list
            Attributes used to train the gaussian process regressor.
        target_cols : list
            Corresponding target values (can be vectors).
        """
        self.attr_cols = attr_cols
        self.target_cols = target_cols

    def element_name(self):
        """Name for this element for logging purposes"""
        return "gp_surrogate"

    def element_version(self):
        """Version of this element for logging purposes"""
        return "0.1"

    def analyse(self, data_frame=None):
        """Perform the basis stats analysis on the input `data_frame`.

        Analysis is based on `pandas.Dataframe.describe` and results in
        values for: count, mean, std, min, max and 25%, 50% & 75% percentiles
        for each value in the analysis.

        The data_frame is grouped according to `self.groupby` if specified and
        analysis is performed on the columns selected in `self.qoi_cols` if set.

        Parameters
        ----------
        data_frame : :obj:`pandas.DataFrame`
            Summary data produced through collation of simulation output.

        Returns
        -------
        :obj:`pandas.DataFrame`
            Basic statistic for selected columns and groupings of data.
        """
        x = data_frame[self.attr_cols].values
        y = data_frame[self.target_cols].values

        gp = GaussianProcessRegressor()
        gp.fit(x, y)
        return gp



from easyvvuq.sampling.sweep import BasicSweep, wrap_iterable
import pytest


@pytest.fixture
def basic_sweep_sampler():
    return BasicSweep({'a': [1, 2, 3], 'b': [4, 5, 6]})


def test_wrap_iterable():
    res = []
    for var_name, val in wrap_iterable('a', [1, 2, 3]):
        res.append((var_name, val))
    assert(res == [('a', 1), ('a', 2), ('a', 3)])


def test_sweep(basic_sweep_sampler):
    res = []
    for run_dict in basic_sweep_sampler:
        res.append(run_dict)
    assert(res == [{'a': 1, 'b': 4}, {'a': 1, 'b': 5}, {'a': 1, 'b': 6},
                   {'a': 2, 'b': 4}, {'a': 2, 'b': 5}, {'a': 2, 'b': 6},
                   {'a': 3, 'b': 4}, {'a': 3, 'b': 5}, {'a': 3, 'b': 6}])


def test_init():
    sweep = BasicSweep({'a': [1, 2, 3], 'b': [4, 5, 6]}, 10)
    with pytest.raises(StopIteration):
        sweep.__next__()


def test_element_version(basic_sweep_sampler):
    assert(isinstance(basic_sweep_sampler.element_version(), str))


def test_is_finite(basic_sweep_sampler):
    assert(basic_sweep_sampler.is_finite())


def test_n_samples(basic_sweep_sampler):
    assert(basic_sweep_sampler.n_samples() == 9)


def test_is_restartable(basic_sweep_sampler):
    assert(basic_sweep_sampler.is_restartable())


def test_get_restart_dict(basic_sweep_sampler):
    restart_dict = basic_sweep_sampler.get_restart_dict()
    assert(restart_dict['sweep'] == {'a': [1, 2, 3], 'b': [4, 5, 6]})
    assert(restart_dict['count'] == 0)


def test_basic_sweep_single_list():
    assert(len(list(BasicSweep({'a': [1, 2, 3]}))) == 3)



import os
import logging
from easyvvuq.constants import Status

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


logger = logging.getLogger(__name__)


class Worker:
    """
    A Worker is a stripped down version of the Campaign, providing access to runs in the database,
    encoding etc. The purpose is, for example, to allow encoding/running/decoding of runs to take
    place in a separate process (script) from the main script.
    """

    def __init__(
            self,
            db_type="sql",
            db_location=None,
            campaign_name=None,
            app_name=None,
            write_to_db=True
    ):

        self.campaign_name = campaign_name
        self.db_type = db_type
        self.db_location = db_location
        self.write_to_db = write_to_db

        if self.db_type == 'sql':
            from .db.sql import CampaignDB
        else:
            message = (f"Invalid 'db_type' {self.db_type}. Supported types "
                       f"are: 'sql'.")
            logger.critical(message)
            raise RuntimeError(message)

        # Open session to the database
        logger.info(f"Opening session with CampaignDB at {self.db_location}")
        self.campaign_db = CampaignDB(location=self.db_location,
                                      new_campaign=False,
                                      name=self.campaign_name)
        self.campaign_id = self.campaign_db.get_campaign_id(self.campaign_name)

        # Resurrect the app encoder and decoder elements
        self._active_app_name = app_name
        self._active_app = self.campaign_db.app(name=app_name)
        (self._active_app_encoder,
         self._active_app_decoder,
         self._active_app_collater) = self.campaign_db.resurrect_app(app_name)

    def encode_runs(self, run_id_list):
        """
        Runs the encoder for all run ids listed in run_id_list.
        """

        # Get the encoder for this app. If none is set, only the directory structure
        # will be created.
        active_encoder = self._active_app_encoder
        if active_encoder is None:
            logger.warning('No encoder set for this app. Creating directory structure only.')

        # Loop through all runs in the run_id_list
        runs_dir = self.campaign_db.runs_dir()

        for run_id in run_id_list:

            run_data = self.campaign_db.run(run_id)

            # Make run directory
            target_dir = os.path.join(runs_dir, run_id)
            os.makedirs(target_dir)
            self.campaign_db.set_dir_for_run(run_id, target_dir)

            if active_encoder is not None:
                active_encoder.encode(params=run_data['params'],
                                      target_dir=target_dir)

        # Update run statuses in db (if worker is authorized to write to DB)
        if self.write_to_db:
            self.campaign_db.set_run_statuses(run_id_list, Status.ENCODED)

    def call_for_each_run(self, fn, status=Status.ENCODED):
        """
        Loop through all runs in this campaign with the specified status,
        and call the specified user function for each.
        """
        for run_id, run_data in self.campaign_db.runs(status=status):
            fn(run_data['run_dir'], run_data['params'])



from .base import BaseSamplingElement
import itertools
import logging
import functools

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"

logger = logging.getLogger(__name__)


def wrap_iterable(var_name, iterable):
    for val in iterable:
        yield (var_name, val)


class BasicSweep(BaseSamplingElement, sampler_name="basic_sweep"):

    def __init__(self, sweep=None, count=0):
        """
            Expects dict of var names, and their corresponding lists of values to cycle through
        """
        self.sweep = sweep

        gens = []
        for var_name, iterable in self.sweep.items():
            gens.append(wrap_iterable(var_name, iterable))

        # Combine all the iterables/generators into one
        self.sweep_iterator = itertools.product(*gens)

        self.count = 0
        for i in range(count):
            try:
                self.__next__()
            except StopIteration:
                logger.warning("BasicSweep constructed, but has no samples left to draw.")

    def element_version(self):
        return "0.1"

    def is_finite(self):
        return True

    def n_samples(self):
        """Returns the number of samples in this sampler.

        Returns
        -------
        a product of the lengths of lists passed to BasicSweep
        """
        return functools.reduce(
            lambda x, y: x * y, [len(lst) for lst in [self.sweep[key] for key in self.sweep]], 1)

    def __next__(self):
        # Will raise StopIteration when there are none left
        sweep_run = self.sweep_iterator.__next__()

        run_dict = {}
        for var_name, value in sweep_run:
            run_dict[var_name] = value

        self.count += 1
        return run_dict

    def is_restartable(self):
        return True

    def get_restart_dict(self):
        return {"sweep": self.sweep, "count": self.count}



import os
from .base import BaseEncoder
import logging

__copyright__ = """

    Copyright 2018 Robin A. Richardson, David W. Wright

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


class DirectoryBuilder(BaseEncoder, encoder_name="directory_builder"):
    """DirectoryBuilder builds the specified directory structure for a Run.

    The dir structure is specified by the 'tree' parameter. This should be a dict of dicts,
    for example:

    tree = {'a' : {'b' : {'c' : None, 'd' : None}}, 'e' : {'f' : None}}


    Parameters
    ----------
        tree    : dict of dicts
            The desired directory structure
    Attributes
    ----------

    """

    def __init__(self, tree):
        self.tree = tree

    def encode(self, params={}, target_dir=''):
        """Builds the directory structure specified in self.tree into the `target_dir` directory

        Parameters
        ----------
        params        : dict
            Parameter information in dictionary.
        target_dir    : str
            Path to directory where application input will be written.
        """

        if not target_dir:
            raise RuntimeError('No target directory specified to encoder')

        self.create_dir_tree(self.tree, target_dir)

    def create_dir_tree(self, dirtree, root):
        # A beautiful Vytas creation
        if dirtree is not None:
            for directory in dirtree.keys():
                os.mkdir(os.path.join(root, directory))
                self.create_dir_tree(dirtree[directory], os.path.join(root, directory))

    def get_restart_dict(self):
        return {"tree": self.tree}

    def element_version(self):
        return "0.1"



import time
import threading

__copyright__ = """

    Copyright 2020 Vytautas Jancauskas

    This file is part of EasyVVUQ

    EasyVVUQ is free software: you can redistribute it and/or modify
    it under the terms of the Lesser GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    EasyVVUQ is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    Lesser GNU General Public License for more details.

    You should have received a copy of the Lesser GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
__license__ = "LGPL"


class ActionStatuses:
    """A class that tracks statuses of a list of actions.

    Parameters
    ----------
    statuses: list of ActionStatus
        a list of action statuses to track
    poll_sleep_time: int
        a time to sleep for after iterating over all active statuses
        before starting again

    """

    def __init__(self, statuses, batch_size, poll_sleep_time=1):
        self.ready = list(statuses)
        self.active = []
        self.finished = []
        self.failed = []
        self.batch_size = batch_size
        self.poll_sleep_time = poll_sleep_time
        self._stats = {
            'ready': len(self.ready),
            'active': 0,
            'finished': 0,
            'failed': 0
        }
        polling_thread = threading.Thread(target=self.poll)
        polling_thread.start()

    def poll(self):
        """Iterate over active actions, finalize finished ones,
        sort finished actions into finished and failed. An action
        is considered finished if it has finished executed and reports
        success. It is considered failed if it has reported failure and
        is considered active (running) otherwise.
        """
        while True:
            ready_ = []
            for status in self.ready:
                if len(self.active) < self.batch_size:
                    status.start()
                    self.active.append(status)
                    self._stats['active'] += 1
                    self._stats['ready'] -= 1
                else:
                    ready_.append(status)
            self.ready = ready_
            active_ = []
            for status in self.active:
                if status.finished():
                    self._stats['active'] -= 1
                    if status.succeeded():
                        self._stats['finished'] += 1
                        status.finalise()
                        self.finished.append(status)
                    else:
                        self._stats['failed'] += 1
                        self.failed.append(status)
                else:
                    active_.append(status)
            self.active = active_
            if (not active_) and (not ready_):
                break
            else:
                time.sleep(self.poll_sleep_time)

    def progress(self):
        """Return the number of active, finished and failed jobs.
        """
        return self._stats



#!/usr/bin/env python3
"""
Analytic anisotropic function to test the SC sampler on
"""

# scalar analytic test function, defined on [0, 1]**d
import json
import sys
import numpy as np


def poly_model(theta):
    """
    Analytic test function where some parameters are more important than others.

    Parameters
    ----------
    theta :  array of input parameters in [0, 1]

    Returns
    -------
    (float) function value
    """
    sol = 1.0
    for i in range(d):
        sol *= 3 * a[i] * theta[i]**2 + 1.0
    return sol / 2**d


# stocastic dimension of the problem
d = 3
a = np.ones(d) * 0.01
# effective dimension of the problem
effective_d = 1
a[0:effective_d] = 1.0

# the json input file containing the values of the parameters, and the
# output file
json_input = sys.argv[1]

with open(json_input, "r") as f:
    inputs = json.load(f)
output_filename = inputs['outfile']

theta = []
for i in range(d):
    theta.append(float(inputs['x' + str(i + 1)]))
theta = np.array(theta)

result = poly_model(theta)

# output csv file
header = 'f'
np.savetxt(output_filename, np.array([result]),
           delimiter=",", comments='',
           header=header)



#!/usr/bin/env python3

# scalar analytic test function, defined on [0, 1]**d
import numpy as np
import json
import sys
import logging


def sobol_g_func(theta, d=5, a=[0.0, 0.5, 3.0, 9.0, 99.0]):
    Y = 1.0
    for i in range(d):
        Y *= (np.abs(4.0 * theta[i] - 2.0) + a[i]) / (1.0 + a[i])
    return Y


if __name__ == '__main__':
    # parameters required by test function
    a = [0.0, 0.5, 3.0, 9.0, 99.0]
    # stocastic dimension of the problem
    d = 5
    # the json input file containing the values of the parameters, and the
    # output file
    json_input = sys.argv[1]
    with open(json_input, "r") as f:
        inputs = json.load(f)
    output_filename = inputs['outfile']
    theta = []
    for i in range(d):
        theta.append(float(inputs['x' + str(i + 1)]))
    theta = np.array(theta)
    result = sobol_g_func(theta)
    logging.debug(result)
    # output csv file
    header = 'f'
    np.savetxt(output_filename, np.array([result]),
               delimiter=",", comments='',
               header=header)



from easyvvuq.sampling import ReplicaSampler
from easyvvuq.sampling import BasicSweep
from easyvvuq.sampling import EmptySampler
import pytest


@pytest.fixture
def replica_sampler():
    return ReplicaSampler(BasicSweep({'a': [1, 2], 'b': [3, 4]}))


def test_infite_exception():
    with pytest.raises(RuntimeError):
        ReplicaSampler(EmptySampler())


def test_is_finite(replica_sampler):
    assert(not replica_sampler.is_finite())


def test_element_version(replica_sampler):
    assert(replica_sampler.element_version() == '0.1')


def test_n_samples(replica_sampler):
    with pytest.raises(RuntimeError):
        replica_sampler.n_samples()


def test_replica_sampler_ensemble(replica_sampler):
    params = next(replica_sampler)
    assert(params == {'a': 1, 'b': 3, 'ensemble': 0})
    params = next(replica_sampler)
    assert(params == {'a': 1, 'b': 4, 'ensemble': 1})
    params = next(replica_sampler)
    assert(params == {'a': 2, 'b': 3, 'ensemble': 2})
    params = next(replica_sampler)
    assert(params == {'a': 2, 'b': 4, 'ensemble': 3})
    params = next(replica_sampler)
    assert(params == {'a': 1, 'b': 3, 'ensemble': 0})
    params = next(replica_sampler)
    assert(params == {'a': 1, 'b': 4, 'ensemble': 1})
